    page 66,132

    include ..\aal5\atm.inc

_DATA   segment word public 'DATA'
    extrn _ip_queue_write:near
    extrn _vc_state:near
_DATA   ends
_BSS    segment word public 'BSS'
_BSS    ends
_TEXT   segment byte public 'CODE'
_TEXT   ends
DGROUP  group   _DATA,_BSS
    assume  cs:_TEXT,ds:DGROUP
_TEXT   segment byte public 'CODE'
    .386p







; conditional compile switches
;
write_to_card       =   0               ; only for debug!!!
read_from_high_mem  =   0               ; only for debug!!!
read_from_card      =   1
write_to_high_mem   =   1
unroll_cell_copy    =   1
look_at_header      =   1
look_at_payload     =   1







;
; void _fastcall aal5_reassemble( Bit32 far *cell, unsigned char count )
;
; count parameter is passed in AL register
;
; cell parameter is passed on stack
; (need to get it in registers in future!)
;
;
;
; extracts IP headers from the given number of cells
;
; only looks at the first DWORD of each 52 byte cell
; to determine whether to copy the cell
;
; if this is the first cell seen on this VCI and
; the VCI is in AAL5 mode, copy it to the IP queue
; WITHOUT using memcpy (16-bit string move instruction)
; which is glacially (10x) slow
;
; we need to deal with wrap within 8K buffer someday???
;
;
;
; register assignments:
;       CH          loop counter
;       CL          shift count for VCI state bits
;       DS:BX       offset into VC state table of this VCI
;       ES:SI       source cells
;       GS:EDI      dest IP headers
;       EDX         ATM header (until we decide what to do w/ cell)
;       EAX         scratch
;       FS          unused
;       high 16 bits of EAX, EBX, ECX, ESI unused
;
@aal5_reassemble$qnuluc proc    near

    ; set up the stack frame
    ;
    push bp
    mov bp, sp

    ; we don't use any registers that the caller
    ; expects to be preserved, so none need to be saved
    ;
    push es
    push edi
    push si

    ; set up the pointers
    ;
    les si, dword ptr [bp + 4]
    mov edi, dword ptr [_ip_queue_write]

    ; set up the flags and jump to the loop conditional
    ;
    mov ch, al
    or ch, ch
    jmp loop_condition

not_user_cell:
    ; we could have coded the conditional jump
    ; which got us here to go directly to loop_counter,
    ; but we want a place for future code and breakpoints
    ;
    jmp loop_counter

loop_top:

if look_at_header
    ; get the cell header
    ;
    mov edx, dword ptr es:[ si ]

    ; swap its bytes so we can look at the VCI
    ; and VPI as whole fields
    ;
    ; we must hand-assemble the "BSWAP EDX" instruction
    ; because MASM 5.10 doesn't understand anything
    ; which is 80486- or Pentium-specific
    ;
    db 66h, 0Fh, 0CAh

    ; make sure that this is a user data cell
    ; (skip it if it is OAM)
    ;
    test edx, mask OAM
    jnz not_user_cell

    ; get the VCI in BX register
    ;
    mov ebx, edx
    and ebx, dword ptr mask VCI
    shr ebx, VCI

    ; calculate the address and shift
    ; for the VCI state table entry
    ;
    mov cl, bl
    shr bx, 2
    and cl, 3
    shl cl, 1

    ; get the state for this VC into AL
    ;
    ; note that we rotate, not shift, so
    ; the bits for other VC's are still
    ; present in the register for later
    ; when we want to update state in table
    ;
    mov al, byte ptr _vc_state[ bx ]
    ror al, cl

    ; if this a non-AAL5 VC, skip the cell
    ;
    test al, VC_IS_AAL5
    jz not_aal5_vc

    ; save the state in AH for use after
    ; we have updated the VC state table
    ;
    mov ah, al

    ; if this is the last cell...
    ;
    test edx, mask LAST
    jz not_aal5_last

    ; forget that this VC saw a cell
    ;
    and al, not VC_SAW_CELL
    jmp update_state_table

not_aal5_last:
    ; otherwise must not be the last cell
    ;

    ; remember that this VC saw a cell already
    ;
    or al, VC_SAW_CELL

update_state_table:
    ; rotate the state bits back into position
    ;
    rol al, cl

    ; save bits for all 4 VC's into table
    ;
    mov byte ptr _vc_state[ bx ], al

    ; if this VC has already seen a cell,
    ; skip to the next cell
    ;
    test ah, VC_SAW_CELL
    jnz loop_pointer

    ;*******************************************
    ; at this point we can reuse EAX, EBX, EDX, CL
    ; because we are done with VC table entry
    ; and cell header
    ;
    ; we only end up using EAX, though
    ;*******************************************
endif   ; look_at_header

    ; now we know we want to copy the cell payload
    ; because this is the first cell in the PDU
    ;

if look_at_payload
    ; skip to the payload
    ;
    add si, 4

    ; copy 12 DWORD's while moving the pointers forward
    ;
    if unroll_cell_copy
        pos = 0
        rept 12
        if write_to_card                    ; debug only
            mov es:[ si + pos ], eax
        endif
        if read_from_card
            mov eax, es:[ si + pos ]
        endif
        if write_to_high_mem
            mov gs:[ edi + pos ], eax
        endif
        if read_from_high_mem               ; debug only
            mov eax, gs:[ edi + pos ]
        endif
        pos = pos + 4
        endm
        add si, 48
        add edi, 48
    else
        mov cl, 12
copy_loop:
        if read_from_card
            mov eax, es:[ si ]
        endif
        if write_to_high_mem
            mov gs:[ edi ], eax
        endif
        add si, 4
        add edi, 4
        dec cl
        jnz copy_loop
    endif

    ; no need to add anything to the cell pointer,
    ; since our individual increments while copying
    ; have cumulatively moved it to the next cell
    ;
    jmp short loop_counter
else
    jmp short loop_pointer
endif   ; look_at_payload

not_aal5_vc:
    ; want some code here as a placeholder for
    ; future code and current breakpoints
    ;
    nop

loop_pointer:
    ; move the cell pointer forward
    ;
    add si, 52

loop_counter:
    ; just decrement the cell counter
    ;
    dec ch

loop_condition:
    ; if there are more cells to do, go back up
    ;
    jne loop_top

    ; flush the IP queue pointer
    ;
    ; maybe we should do this inside the loop for safety???
    ;
    mov dword ptr _ip_queue_write, edi

    ; restore the caller's non-scratch registers
    ;
    pop si
    pop edi
    pop es

    ; clean up the stack frame
    ;
    mov sp, bp
    pop bp

    ; take the caller's parameters off the stack
    ; and return to him
    ;
    ret 4

@aal5_reassemble$qnuluc endp




public @aal5_reassemble$qnuluc




_TEXT ends

    end
