#ifndef _POINT_DRIVER_H
#define _POINT_DRIVER_H



// conditional compile switches that our clients use
//
#define CARD_INIT_PROVIDES_BUFFERS      TRUE



/* header files we depend upon
 */
#include "../apptel/point.h"            // base class AtmCard
#include "../intel/i8259a.h"            // irq_to_intr() class InterruptController
#include "../integrat/buffer.h"         // class BufferBlock



/* errors
 */
#define ERROR_POINT2_THUNK_ALLOC                300
#define ERROR_POINT2_BUFFERS_EXHAUSTED          301     // not used
#define ERROR_POINT2_CANNOT_TX                  302
#define ERROR_POINT2_MUST_RX                    303



class AtmCardRunning : public AtmCard
{

public:
    unsigned irq_level;
    unsigned char interrupt_number;
    Bit48 old_handler;
    vfp thunk;
    BufferBlock *my_block;
    static InterruptController irq;
    enum InitType
    {
        NormalAtm,
        NormalPpp,
        PassIdleCells
    };

private:
    static int card_count;
    int card_was_initialized;
    AmccInterruptControlStatus_register amcc_intr;
    AmccBusMasterControlStatus_register amcc_bus;
    unsigned interrupt_count;
    int interrupt_result;

public:
    // constructor
    //
    AtmCardRunning( Bit8 card_index0 );

    // destructor
    //
    ~AtmCardRunning();

    // gets the card running the way you want
    //
    int card_init( InitType itype, LoopbackSelection ltype, int want_rx, int want_tx );

    unsigned get_interrupt_count( void )
    {
        return interrupt_count;
    }

    int get_interrupt_result( void )
    {
        int result = interrupt_result;
        interrupt_result = 0;
        return result;
    }

    // no need to implement this for Apptel POINT cards
    //
    void timer_handler( void )
    {
    }

    // maybe later implement sucking the path trace message,
    // which some routers/switches use to identify themselves
    // with an ASCII string, from the SUNI 622's buffer
    //
    void tick( void ) { }

    // not implemented in hardware yet
    //
    int start_clearing_timestamp( void ) { return 0; }
    int wait_for_timestamp_to_clear( void ) { return 0; }

    // called when BufferQueue free list stops being empty
    // in case this card was in need of a buffer
    //
    void foreground_replenish( void );

    // called when BufferQueue sendable list stops being empty
    // for this card in particular, in case this card was
    // in need of a buffer
    //
    void foreground_tx_replenish( void )
    {
    }

    void set_pending_block_usage( void );

private:
    // static interrupt handler that we can take the address of
    //
    static int interrupt_handler0( AtmCardRunning *this0,
        Bit32 ebp, Bit32 edi, Bit32 esi,
        Bit32 ds, Bit32 es, Bit32 edx, Bit32 ecx, Bit32 ebx, Bit32 eax,
        Bit32 eip, Bit32 cs, Bit32 flags
    )
    {
        return this0->interrupt_handler();
    }

    int interrupt_handler( void );
    void interrupt_init( void );
    void interrupt_exit( void );

    int write_completed_handler( void );

    int card_init_suni( InitType type );
    void card_exit( void );
};



#endif  /* not included yet */
