#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/nvram.h"            // nvram_init_stats() nvram_print_stats() nvram_read_or_write()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    Bit16 nvram_address;
    Bit8 nvram_data;
    int want_write;
    int result;
    const char *program_name = argv[0];
    AtmCard *card;

    // parse command-line arguments
    //
    if( argc<3 || argc>4 )
    {
        fprintf( stderr, "reads or writes single NVRAM bytes on the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex address {new-value}\n", program_name );
        return ERROR_NVRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    nvram_address = strtoul( argv[2], NULL, 0 );

    if( argc == 4 )
    {
        nvram_data = strtoul( argv[3], NULL, 0 );
        want_write = TRUE;
    }
    else
        want_write = FALSE;

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_NVRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    nvram_init_stats();

    result = nvram_read_or_write( card->amcc, nvram_address, want_write, &nvram_data );

    nvram_print_stats();

    delete card;

    if( result )
    {
        fprintf( stderr, "error %d from nvram_read_or_write()\n", result );
        return result;
    }

    if( want_write )
        printf( "wrote new value 0x%.2X at address 0x%.4X\n", nvram_data, nvram_address );
    else
        printf( "read old value 0x%.2X at address 0x%.4X\n", nvram_data, nvram_address );

    return 0;
}
