#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/sram.h"             // sram_init_stats() sram_print_stats() sram_read_or_write()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    Bit32 sram_address;
    Bit32 sram_length;
    Bit32 count;
    Bit32 sram_data;
    int result;
    unsigned line;
    const char *program_name = argv[0];
    AtmCard *card;

    // parse command-line arguments
    //
    if( argc != 4 )
    {
        fprintf( stderr, "dumps a range of SRAM words from the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex address length-in-words\n", program_name );
        return ERROR_SRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    sram_address = strtoul( argv[2], NULL, 0 );
    sram_length = strtoul( argv[3], NULL, 0 );

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_SRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    sram_init_stats();

    printf( "dumping POINT SRAM values starting at address 0x%.8X, word length 0x%.8X\n", sram_address, sram_length );

    for( count=0; count<sram_length; count+=8 )
    {
        // this line assumes SRAM addresses have 20 bits or less
        //
        printf( "%.5X  ", sram_address+count );

        for( line=0; line<8 && count+line<sram_length; line++ )
        {
            result = sram_read_or_write( &card->both_fpgas->Receive, sram_address+count+line, FALSE, &sram_data );
            if( result )
                break;
            printf( " %.8X", sram_data );
        }

        printf( "\n" );
    }
    printf( "\n" );

    sram_print_stats();

    delete card;

    if( result )
    {
        fprintf( stderr, "error %d from sram_read_or_write() on %uth word\n", result, count );
        return result;
    }

    return 0;
}
