#ifndef _REALMODE_H
#define _REALMODE_H

// headers we depend upon
//
#include "..\types\types.h"

#ifdef __cplusplus
extern "C" {
#endif

// pointer to a protected mode C function that is called from real mode
// and returns a pointer to altered real mode registers
//
typedef Bit48 (*cfp)( Bit48 real_regs_pointer, Bit48 real_stack_pointer );

union _Reg32
{
    Bit8 low8;
    Bit16 low16;
    Bit32 all;
};
typedef union _Reg32 Reg32;

struct RealModeRegs
{
    Reg32 edi;
    Reg32 esi;
    Reg32 ebp;
    Bit32 reserved;
    Reg32 ebx;
    Reg32 edx;
    Reg32 ecx;
    Reg32 eax;
    Bit16 flags;
    Bit16 es;
    Bit16 ds;
    Bit16 fs;
    Bit16 gs;
    Bit16 ip;
    Bit16 cs;
    Bit16 sp;
    Bit16 ss;
};

struct RealModeRegsC
{
    Bit48 saved_stack;
    Bit8 *stack;
    cfp function;
    struct RealModeRegs regs;
};

Bit32 real_to_physical( Bit32 segment_offset );

#define MAX_REAL_ADDRESS 0x10FFEFUL
Bit32 physical_to_real( Bit32 physical );

Bit16 alloc_dos_mem( Bit32 length_wanted, Bit32 *length_available,
    Bit16 *selector, Bit16 *segment );

Bit16 realloc_dos_mem( Bit32 length_wanted, Bit32 *length_available,
    Bit16 *selector, Bit16 *segment );

Bit16 free_dos_mem( Bit16 selector );

Bit16 alloc_real_callback( cfp protmode_c_function, Bit8 *protmode_c_stack,
    struct RealModeRegsC *real_regs_c, Bit32 *real_segment_offset );

Bit16 free_real_callback( Bit32 real_segment_offset );

Bit16 simulate_real_interrupt( Bit8 interrupt_number, struct RealModeRegs *real_regs );

#ifdef __cplusplus
}
#endif

#endif  // not included yet
