#ifndef _NETINET_TCP_H
#define _NETINET_TCP_H

typedef u_int   tcp_seq;

/*
 * TCP header.
 * Per RFC 793, September, 1981.
 */

/* part of TCP header that fits in first cell
 */
struct tcphdr_1 {
    u_short th_sport;       /* source port */
    u_short th_dport;       /* destination port */
};

/* part of TCP header that fits in first cell if
 * gigaswitch encapsulation isn't on at the same
 * time as LLC/SNAP encapsulation
 *
 * but it gets really nasty if gigaswitch encapsulation
 * is on at the same time as NLPID encapsulation,
 * since th_ack will straddle 2 cells
 */
struct tcphdr_2 {
    tcp_seq th_seq;         /* sequence number */
    tcp_seq th_ack;         /* acknowledgement number */
};

/* part of TCP header that fits in first cell
 * only if gigaswitch encapsulation is off
 */
struct tcphdr_3 {
#if HOST_IS_BIG_ENDIAN
    u_char  th_off:4;       /* data offset in 32-bit words (TCP header length) */
    u_char  th_x2 :4;       /* (unused) */
#else
    u_char  th_x2 :4;       /* (unused) */
    u_char  th_off:4;       /* data offset in 32-bit words (TCP header length) */
#endif
    u_char  th_flags;
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
    u_short th_win;         /* window */
    u_short th_sum;         /* checksum */
    u_short th_urp;         /* urgent pointer */
};



struct tcphdr {
    struct tcphdr_1 from_cell_1;
    struct tcphdr_2 from_cell_2;
    struct tcphdr_3 from_cell_3;
};

#define TCPOPT_EOL      0           /* must be last option (only this 1 byte) */
#define TCPOPT_NOP      1           /* used for padding (only this 1 byte) */
#define TCPOPT_MAXSEG   2           /* RFC 1323 maximum segment size (len + 2 bytes) option */
#define TCPOPT_WINDOW   3           /* RFC 1323 window scale option (len + 1 byte) */
#define TCPOPT_SACKPERM 4           /* RFC 2018 selective-acknowledge permitted option (len only) */
#define TCPOPT_SACK     5           /* RFC 2018 selective-acknowledge option (len + multiple left/right 32-bit edges) */
#define TCPOPT_TSTAMP   8           /* RFC 1323 timestamp option (len + one each TX/RX 32-bit times) */
#define TCPOPT_CC       11          /* RFC 1644 T/TCP connection count, sent with all segment kinds */
#define TCPOPT_CC_NEW   12          /* RFC 1644 T/TCP new connection count, sent instead of CC during SYN */
#define TCPOPT_CC_ECHO  13          /* RFC 1644 T/TCP echoed connection count, sent with CC during SYN ACK */

/*
 * Default maximum segment size for TCP.
 * For RFC1122 MUST conformance, this needs to be 536.
 */
#define TCP_MSS 536

#define TCP_MAXWIN  65535       /* largest value for (unscaled) window */

#define TCP_MAX_WINSHIFT      14    /* maximum window shift */

/*
 * User options (used with set/getockopt).
 */
//#define TCP_NODELAY 0x01    /* don't delay send to coalesce packets */
#define TCP_MAXSEG  0x02    /* maximum segment size */
/*
 * Number of retransmissions before dropping connection. Larger than
 * TCP_MAXRXTSHIFT or -1 for infinity
 */
#define TCP_RPTR2RXT    0x03    /* set repeat count for R2 RXT timer */

#endif
