// this source file contains substitutes for the functions
// in FS2FLOWS.C that are called by TRACEFLO.CPP
// so we can perform our own analysis
//



// force these variables so this executable will always
// operate on an IP header trace, regardless how CAPTURE.EXE
// is compiled (as determined by CONSTANT.H)
//
// this is only safe to do if this is the only source file
// in this executable which cares about tracerec structure
//
#define CAPTURE_TO_DISK             FALSE
#define CARD_COPIES_ATM_PAYLOAD     TRUE



#include "traceflo.cpp"



// conditional compilation switches
//
#define PRINT_LEADING_ZEROES            FALSE
#define PRINT_LENGTHS_WITH_NO_PACKETS   FALSE
#define WANT_TIMES                      FALSE
#define CHECK_MAC_LENGTH                TRUE
#define DEBUG_MAC_LENGTH                FALSE
#define CHECK_MAC_CONTROL               TRUE
#define DEBUG_MAC_CONTROL               TRUE



// manifest constants
//
#define QUERY_TCP_PORT             22   // use secure shell (ssh) port to get past firewall
//#define QUERY_TCP_PORT           23   // use telnet port for now



// global variables
//
#if WANT_TIMES
    char outfile[] = "times.txt";
    FILE *out_handle;
#endif
#if LINK_HAS_LLC_SNAP
    Bit32 unknown_llc_packets = 0;
    Bit32 unknown_llc_and_non_ip_ethertype_packets = 0;
    Bit32 unknown_oui_packets = 0;
    Bit32 unknown_oui_and_non_ip_ethertype_packets = 0;
#endif
#if LINK_HAS_LLC_SNAP || LINK_HAS_NLPID
    Bit32 non_ip_ethertype_packets = 0;
#endif
#if CHECK_MAC_LENGTH || CHECK_MAC_CONTROL
    Bit32 mac_trailer_missing_packets = 0;
#endif
#if CHECK_MAC_LENGTH
    Bit32 mac_length_lt_ip_packets = 0;
    Bit32 mac_length_gt_ip_packets = 0;
#endif
#if CHECK_MAC_CONTROL
    Bit32 mac_control_ne_0_packets = 0;
#endif
Bit32 ip_length_usage[ IP_MAXPACKET + 1 ];
Bit32 valid_packets = 0;
Bit32 ip_optioned_packets = 0;
Bit32 ip_unknown_flagged_packets = 0;
Bit32 ip_df_flagged_packets = 0;
Bit32 ip_whole_packets = 0;
Bit32 first_ip_fragments = 0;
Bit32 non_first_ip_fragments = 0;
tcp_Socket sock;



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

// for aal5.c to call
//
int flows_per_packet( cellhashrec *packet, Bit8 interface )
{
#if LINK_HAS_LLC_SNAP
    // if LLC is bad...
    //
    if( packet->tr_lsap != LSAP_SNAP
    ||  packet->tr_ctl != CTL_SNAP )
    {
        unknown_llc_packets++;
        if( packet->tr_etype != ETHERTYPE_IP_SWAPPED )
            unknown_llc_and_non_ip_ethertype_packets++;
        return 0;
    }

    // if OUI is bad...
    //
    if( packet->tr_oui[0] != 0
    ||  packet->tr_oui[1] != 0
    ||  packet->tr_oui[2] != 0 )
    {
        unknown_oui_packets++;
        if( packet->tr_etype != ETHERTYPE_IP_SWAPPED )
            unknown_oui_and_non_ip_ethertype_packets++;
        return 0;
    }
#endif

#if LINK_HAS_LLC_SNAP || LINK_HAS_NLPID
    // if ethertype is not IP...
    //
    if( packet->tr_etype != ETHERTYPE_IP_SWAPPED )
    {
        non_ip_ethertype_packets++;
        return 0;
    }
#endif

    // packet is good as far as we are concerned
    //
    // note that flows analysis will not deal with
    // packets with IP options or which are not first
    // of fragment because TCP header has shifted
    // out of the cell, and it doesn't want to bother
    // re-laying out the structure to get ICMP payload
    //
    // also note that flows analysis incorrectly calculates
    // length for IP fragmented packets, since IP header's
    // length in first fragment includes only that fragment,
    // and it doesn't count length of non-first fragments
    //
    valid_packets++;

    // if this packet used IP options...
    //
    if( packet->tr_iphl != (sizeof(ip) >> 2) )
    {
        ip_optioned_packets++;
    }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
    // swap the bytes of the length in the IP header
    //
    packet->tr_len = ntohs( packet->tr_len );

    // swap the bytes of the IP fragment offset
    //
    packet->tr_offset = ntohs( packet->tr_offset );
#endif

    // note that flows analysis would not deal with
    // fragments because TCP header has been shifted
    // and it doesn't want to recalculate field
    // locations
    //

    // if this is a followon fragment...
    //
    if( (packet->tr_offset & IP_OFF_MASK) != 0 )
    {
        non_first_ip_fragments++;
    }

    // otherwise, if more fragments will follow...
    //
    else if( packet->tr_offset & IP_MF )
    {
        first_ip_fragments++;
    }

    // otherwise this must be all of an unfragmented packet
    //
    else
    {
        // if this packet used other IP flags...
        //
        if( packet->tr_offset & IP_UF )
            ip_unknown_flagged_packets++;
        else if( packet->tr_offset & IP_DF )
            ip_df_flagged_packets++;
        else
            ip_whole_packets++;
    }

    // update the count of packets with this length
    //
    ip_length_usage[ packet->tr_len ]++;

#if CHECK_MAC_LENGTH || CHECK_MAC_TRAILER
    // if the last cell in the packet got captured...
    //
    if( packet->tr_gotlast )
    {
#if CHECK_MAC_LENGTH
#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the bytes of the length in the AAL5 trailer
        //
        packet->tr_maclen = ntohs( packet->tr_maclen );
#endif
#if DEBUG_MAC_LENGTH
        fprintf( stdout, "iface %d packet maclen=%d iplen=%d overhead=%d\n",
            interface, packet->tr_maclen, packet->tr_len, MAC_TO_IP_OFFSET );
#endif

        if( packet->tr_maclen > packet->tr_len + MAC_TO_IP_OFFSET )
            mac_length_gt_ip_packets++;
        else if( packet->tr_maclen < packet->tr_len + MAC_TO_IP_OFFSET )
            mac_length_lt_ip_packets++;
#endif      // CHECK_MAC_LENGTH
#if CHECK_MAC_CONTROL
#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the bytes of the control field in the AAL5 trailer
        //
        packet->tr_macctl = ntohs( packet->tr_macctl );
#endif

        if( packet->tr_macctl != 0 )
        {
            mac_control_ne_0_packets++;
#if DEBUG_MAC_CONTROL
            fprintf( stdout, "iface %d packet macctl=%d\n", interface, packet->tr_macctl );
#endif
        }
#endif      // CHECK_MAC_CONTROL
    }
    else
        mac_trailer_missing_packets++;
#endif

#if WANT_TIMES
    // print the timestamp and length in hex to stdout
    //
#if PRINT_LEADING_ZEROES
    fprintf( out_handle, "%d %.8lX%.8lX %.4X\n", interface, packet->tr_when.high, packet->tr_when.low, packet->tr_len );
#else
    if( packet->tr_when.high > 0 )
        fprintf( out_handle, "%d %lX%.8lX %X\n", interface, packet->tr_when.high, packet->tr_when.low, packet->tr_len );
    else
        fprintf( out_handle, "%d %lX %X\n", interface, packet->tr_when.low, packet->tr_len );
#endif
#endif

    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



int flows_per_block( BufferBlock *block )
{
    Bit32 count;
    tracerec *cell;
    int result;

    // for each record in the buffer...
    //
    for( count=0; count<block->used; count++ )
    {

        // shortcut for code that follows
        //
        cell = (tracerec *) (block->buffers + count*BUFFER_SIZE);

        result = cells_reassemble( cell, block->cardindex );
        if( result )
        {
            fprintf( stderr, "error %d from cells_reassemble()\n", result );
            return result;
        }
    }

    return 0;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

int flows_init( int argc, char *argv[], PentiumClock *keystroke_interval,
    PentiumClock *query_interval, PentiumClock *active_interval,
    int *want_facility_loopback, int *want_rx, int *want_tx )
{
    fprintf( stderr, "extracts the timestamp and IP length in ASCII hex from an OC3MON trace file\n" );
    fprintf( stderr, "and counts usage of each IP length\n\n" );

#if WANT_TIMES
    out_handle = fopen( outfile, "w" );
    if( out_handle == NULL )
    {
        fprintf( stderr, "error %d opening output file\n", errno );
        return 3;
    }
#endif

    // initialize packet length histogram to empty
    //
    memset( ip_length_usage, 0, sizeof(ip_length_usage) );

    // create a socket and listen on it
    //
    tcp_open( &sock, QUERY_TCP_PORT );
    tcp_listen( &sock, 0, 0, NULL, 0 );

    // initialize the AAL5 reassembly module
    //
    return cells_init();
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par



int flows_cleanup( void )
{
    int result;
    Bit32 count;

    // flush the last cells out of their hashrecs
    //
    result = cells_cleanup();
    if( result )
    {
        fprintf( stderr, "error %d from cells_cleanup()\n", result );
        return result;
    }

    // show statistics
    //
#if LINK_HAS_LLC_SNAP
    fprintf( stdout, "%lu packets with unknown LLC (not 0xAA 0xAA 0x03) found\n", unknown_llc_packets );
    fprintf( stdout, "%lu of unknown LLC packets also had non IP ethertype\n", unknown_llc_and_non_ip_ethertype_packets );
    fprintf( stdout, "%lu packets with unknown SNAP OUI (not 0x00 0x00 0x00) found\n", unknown_oui_packets );
    fprintf( stdout, "%lu of unknown SNAP OUI packets also had non IP ethertype\n", unknown_oui_and_non_ip_ethertype_packets );
#endif
#if LINK_HAS_LLC_SNAP || LINK_HAS_NLPID
    fprintf( stdout, "%lu packets with non IP (not 0x%.4X) ethertype found\n", non_ip_ethertype_packets, ETHERTYPE_IP );
#else
    fprintf( stdout, "all packets were assumed to be IP without either LLC/SNAP or NLPID encapsulation\n" );
#endif
    fprintf( stdout, "%lu valid packets found\n", valid_packets );
    fprintf( stdout, "%lu of valid packets used IP options\n", ip_optioned_packets );
    fprintf( stdout, "%lu of valid packets were whole and unflagged\n", ip_whole_packets );
    fprintf( stdout, "%lu of valid packets were whole and used unknown IP flag\n", ip_unknown_flagged_packets );
    fprintf( stdout, "%lu of valid packets were whole and asked not to be fragmented\n", ip_df_flagged_packets );
    fprintf( stdout, "%lu of valid packets were first of fragment\n", first_ip_fragments );
    fprintf( stdout, "%lu of valid packets were not first of fragment\n", non_first_ip_fragments );
#if CHECK_MAC_LENGTH || CHECK_MAC_CONTROL
    fprintf( stdout, "%lu of valid packets didn't capture MAC trailer\n", mac_trailer_missing_packets );
#endif
#if CHECK_MAC_LENGTH
    fprintf( stdout, "%lu of trailered packets had MAC length < IP length + %d byte overhead\n", mac_length_lt_ip_packets, MAC_TO_IP_OFFSET );
    fprintf( stdout, "%lu of trailered packets had MAC length > IP length + %d byte overhead\n", mac_length_gt_ip_packets, MAC_TO_IP_OFFSET );
#endif
#if CHECK_MAC_CONTROL
    fprintf( stdout, "%lu of trailered packets had MAC control != 0\n", mac_control_ne_0_packets );
#endif

    printf( "number (in decimal) of packets for each IP length (in decimal):\n" );
    for( count=0; count<=IP_MAXPACKET; count++ )    // assumes count has >16 bits
    {
#if !PRINT_LENGTHS_WITH_NO_PACKETS
        if( ip_length_usage[count] > 0 )
#endif
            fprintf( stdout, "%u %lu\n", count, ip_length_usage[count] );
    }

#if WANT_TIMES
    fclose( out_handle );
#endif

    return 0;
}



int flows_is_query_wanted( void )
{
    if( !tcp_tick( &sock ) )
        return FALSE;

    return sock_established( &sock );
}


int flows_is_query_pending( void )
{
    return _chk_socket( &sock );
}



int flows_query( const char *extra_msg )
{
    cells_query( &sock );

    sock_printf( &sock, "%s\n", extra_msg );

    // close the TCP/IP connection from this end
    //
    // when caller closes or we time out, we will know
    //
    sock_close( &sock );

    return 0;
}
