/* 1005, Tue 1 Jun 99

   FLOWHASH.C:  RTFM Meter, using hash tables for counts

   Copyright (C) 1992-1999 by Nevil Brownlee,
   ITSS Technology Development,  The University of Auckland */

/*
 * $Log: flowhash.c,v $
 * Revision 1.1.1.1.2.17  1999/09/29 22:29:14  nevil
 * Changes (mainly changing // to /* comments) for Irix
 *
 * Revision 1.1.1.1.2.16  1999/09/29 02:42:47  nevil
 * Fix problems discovered on PC meters
 * - Use NETRAMET compile-time variable to leave out C++ declarations
 * - Clean up mistakes in porting OCX_BGP code from meter_pc to meter_oc
 *
 * Revision 1.1.1.1.2.15  1999/09/24 02:58:37  nevil
 * Polish up code to get rid of warning messages from Borland (DOS) compiler.
 * Make manager PeerAddress buffers NSAP_ADDR_LEN bytes long.
 * Add asn_lookup variable - only call FindSubnet if ruleset uses ASNs.
 *
 * Revision 1.1.1.1.2.14  1999/09/22 05:38:40  nevil
 * Improve code to work properly on 64-bit machines
 * - Add OS=ALPHA handling to configure.in
 * - Clean up the Alpha compiler warnings
 * - Change all the snmp-related code to use Bit32 instead of unsigned long
 *
 * Revision 1.1.1.1.2.13  1999/09/14 00:45:26  nevil * 4.3 Release ..
 *  - Implement -D option to run NeTraMet as a daemon
 *  - Tidy up the on-line help displays
 *
 * Revision 1.1.1.1.2.12  1999/05/26 02:41:37  nevil
 * Integrate V6 and ASN code into PC versions of the meter.
 * This required a rework of the makefiles, using @cflags.opt files
 * to provide a much longer command line to the Borland C compiler.
 *
 * Revision 1.1.1.1.2.11  1999/05/18 03:36:26  nevil
 * Implement IPv6 in NeTraMet, and its manager/collectors.
 * - This is controlled by the V6 #define
 * - NeTraMet recognises v6 packets and fishes through their extension
 *     headers until it finds the actual payload.
 * - NeMaC et al display v6 addresses in the fom specified by RFC 2373
 * - fd_util and fd_extract allow colons in addresses (for defining tags)
 *
 * Revision 1.1.1.1.2.10  1999/05/17 00:11:53  nevil
 * Fixed 'reuse of ruleset row' problem:
 *   When a ruleset row is destroyed, set_RowStatus calls free_rulespace()
 *   and recover_flows().  free_rulespace() zeroed its varibales in the
 *   ri[] row, but recover_flows() didn't (it should have zeroed
 *   ri_flow_chain).  This left ri_flow_chain pointing to a free flow,
 *   which cause problems if a manager tried to reuse this row for a
 *   new ruleset.  Fixed by zeroing all of the ri[] row. (Also commented
 *   out the zeroing code from free_rulesace).
 *
 * Revision 1.1.1.1.2.9  1999/04/13 05:13:33  nevil
 * Use log_msg for ALL error messages
 *
 * Revision 1.1.1.1.2.8  1999/02/16 03:42:17  nevil
 * *** empty log message ***
 *
 * Revision 1.1.1.1.2.7  1999/02/15 21:24:06  nevil
 * Distribution file for 4.3b9
 *
 * Revision 1.1.1.1.2.5  1999/01/20 03:55:49  nevil
 * Implementation of TCP attributes, part 4
 *
 * Revision 1.1.1.1.2.4  1999/01/08 01:31:52  nevil
 * Implementation of TCP attributes, part 3
 *
 * Revision 1.1.1.1.2.3  1998/12/22 23:53:07  nevil
 * Implementation of TCP attributes, part 2
 *
 * Revision 1.1.1.1  1998/11/16 03:57:28  nevil
 * Import of NeTraMet 4.3b3
 *
 * Revision 1.1.1.1.2.2  1998/11/16 02:46:57  nevil
 * Implementation of TCP attributes, part 1
 *
 * Revision 1.1.1.1.2.1  1998/11/11 23:14:43  nevil
 * Only include malloc.h if we HAVE_MALLOC_H
 *
 * Revision 1.1.1.1  1998/10/28 20:31:27  nevil
 * Import of NeTraMet 4.3b1
 *
 * Revision 1.1.3.1.2.3  1998/10/27 04:09:01  nevil
 * Modified OCX defines for 'flat memory' testing for 32-bit DOS
 * meters.  Changed some data defines to allow p2primes[] comparisons
 * to work  properly for 16-bit DOD meter
 *
 * Revision 1.1.3.1.2.2  1998/10/19 22:32:44  nevil
 * Meter improvements, mostly arising from developments for the
 * OCxMON meter.  These are documented in notes_oc.txt
 *
 * Revision 1.1.3.1.2.1  1998/10/18 20:51:17  nevil
 * Added Nicolai's patches, some 'tidying up' of the source
 *
 * Revision 1.1.3.1  1998/10/13 02:48:35  nevil
 * Import of Nicolai's 4.2.2
 *
 * Revision 1.6  1998/07/21 00:40:14  rtfm
 * Implement popto/poptoact actions for SRL
 *
 * Revision 1.3  1998/05/22 03:57:35  rtfm
 * Implement better hashing algorithm for flow and rule tables
 */

#define DEBUG  /* Required if any of the following are set */

#define PME_TRACE      1  /* Trace rules during pattern match */
#define PM_DEBUG       1
#define DIST_DBG       0
#define RS_CH_DEBUG    0
#define V6_TESTV4      0  /* Turns IP4 addresses into ::IP4 addresses */
#define noSTACK_CHECK  /* Keep an eye on pattern & return stack extents */
#define noGC_TEST  /* Monitor garbage collector */

#define noMATCH_TRACE  /* Testing, testing .. */
#define noCOMPARE_TRACE
#define noCOUNT_TRACE
#define noP_TRACE
#define noGC_DEBUG
#define noTESTING
#define noGC_TEST1
#define noH_TEST
#define TCP_TESTING  0
#define TCP_TRACE    0

#define noDECNET

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#if NEW_ATR
#include <math.h>
#endif

#if !defined(DOS)
# ifdef HAVE_MALLOC_H
# include <malloc.h>
# endif
#include <sys/time.h>
#include <netinet/in.h>
#include "../bgp/types/types.h"  /* Needed here for subnet.h */

#else

#include <alloc.h>
#include <mem.h>
#include <conio.h>
#include "..\intel\byteswap.h"
# ifdef DEBUG
# include <dos.h>
# endif
#endif  /* DOS */

#include "ausnmp.h"

#include "asn1.h"
#include "snmp.h"
#include "snmpimpl.h"

#include <sys/types.h>
#include <string.h>

#include "pktsnap.h"

#define EXTFLOW
#include "flowhash.h"
#include "decnet.h"
#include "met_vars.h"

#if TCP_PKT_TRACE

# define TRACE_SEQ_DECR  0
# define TRACE_NEG_BLP   0
# define BLP_FRACT  0.50

FILE *pt_file;
# define ptf_title "ntm.trace"
void print_pkt_trace(FILE *f,  char *id,  /* Identifies invocation */
   struct subflow *sfp, char *fmt, ...);  /* Supporting detail */
#endif

#ifdef DEBUG
void quack(int x) { x; }  /* Use for a debugger break point */

void paddr(Bit8 huge *a);
void daddr(Bit8 huge *a);
void dport(Bit8 huge *a);
void fdaddr(FILE *f, Bit8 huge *a);
void fdport(FILE *f, Bit8 huge *a);
void pkey(char *msg, struct flow huge *n);
#endif

#if PM_DEBUG
int pm_dbg = 0;
char dbg_msg[80];
#endif


#if NF_OCX_BGP
# if defined(__unix__)  /* Unix meter */
#  include "../bgp/integrat/readbgp.h"
#  include "../bgp/integrat/subnet.h"
# else  /* OCX or DOS */
#  include "..\integrat\readbgp.h"
#  include "..\integrat\subnet.h"
# endif
#endif

#if FLAT_MEM  /* Use flat memory for flows */
# if OC3_NTM  /* Testing on 4.1 NeTraMet */
#   include "..\integrat\flat.h"  /* flat_alloc() flat_remaining() */
# else
#   include "..\dpmi32\flat.h"    /* flat_alloc() flat_remaining() */
# endif
#endif

void free_flow(struct flow huge *q)  /* Free memory alloc'ed for a flow */
       /* Caution: must call mark_idle() before calling free_flow() <<< */
{
#if NEW_ATR
   struct distribution *d, *nd;
   for (d = q->distrib_list; d != NULL; d = nd) {
      nd = d->next;
      free_dist(d);
      }
#endif
#if TCP_ATR
   if (q->sfdata != NULL) {
      terminate_all_subflows(q->sfdata);
      free_sf_data(q->sfdata);
      }
#endif
/*   q->FlowRuleSet = 0;  /* Now it doesn't look like part of an rs chain! */
   /* Don't need this, it's done in mark_idle().  Nevil, 15 May 99 */
   }

#if NEW_ATR
static int complained_about_distribs = 0;
struct distribution *get_dist(void)  /* Get an unused distribution */
{
   struct distribution *d;
   if (free_distribs != NULL) {
      d = free_distribs;  free_distribs = free_distribs->next;
      --n_free_distribs;
      return d;
      }
   else {
      if (!complained_about_distribs) {
         log_msg(LOG_WARNING, 0, "Not enough distributions <<<\n");
         complained_about_distribs = 1;
         }
      return NULL;
      }
   }

void free_dist(struct distribution *d)  /* Return distribution to free list */
{
   switch (d->selector) {
   case FTTOBITRATE:
   case FTFROMBITRATE:
   case FTTOPDURATE:
   case FTFROMPDURATE:
      unlink_distrib_from_event(d);
      break;
      }
   d->next = free_distribs;  free_distribs = d;
   ++n_free_distribs;
   complained_about_distribs = 0;
   }

static int complained_about_events = 0;
struct event *get_event(void)  /* Get an unused event */
{
   struct event *e;
   if (free_events != NULL) {
      e = free_events;  free_events = free_events->next;
      --n_free_events;
      return e;
      }
   else {
      if (!complained_about_events) {
         log_msg(LOG_WARNING, 0, "Not enough dist events <<<\n");
         complained_about_events = 1;
         }
      return NULL;
      }
   }

void free_event(struct event *e)  /* Return event to free list */
{
   e->next = free_events;  free_events = e;
   ++n_free_events;
   complained_about_events = 0;
   }

void schedule_event(struct event *ev)
{
   struct event *eqp, *lqp;
   lqp = &eventq;  eqp = eventq.next;
   for ( ; eqp != NULL; lqp = eqp, eqp = eqp->next) {
      if (ev->t <= eqp->t) break;
      }
   lqp->next = ev;  ev->next = eqp;
   return;
   }

int make_distrib_list(struct flow huge *f);
int transform(struct distribution *d, double y);
Bit32 scale(struct distribution *d, double y);

void compute_rate(struct distribution *d)
{
   struct flow huge *f;
   counter64 r;
   double delta;
   f = d->fp;
   switch (d->selector) {
   case FTTOBITRATE:
      subtr64(r, f->UpOctets,f->LastUpOctets);
      assign64(f->LastUpOctets, f->UpOctets);
      delta = (doublefrom64(r))*8.0;  /* Bits */
      break;
   case FTFROMBITRATE:
      subtr64(r, f->DownOctets,f->LastDownOctets);
      assign64(f->LastDownOctets, f->DownOctets);
      delta = (doublefrom64(r))*8.0;  /* Bits */
      break;
   case FTTOPDURATE:
      subtr64(r, f->UpPDUs,f->LastUpPDUs);
      assign64(f->LastUpPDUs, f->UpPDUs);
      delta = doublefrom64(r);  /* Packets */
      break;
   case FTFROMPDURATE:
      subtr64(r, f->DownPDUs,f->LastDownPDUs);
      assign64(f->LastDownPDUs, f->DownPDUs);
      delta = doublefrom64(r);  /* Packets */
      break;
      }
   ++d->counts[transform(d, delta/((double)d->Parameter1))];
   }

void check_rates(Bit32 now)
{
   struct event *eqp;
   struct distribution *d;
   for (;;) {
      eqp = eventq.next;
      if (eqp == NULL || eqp->t > now) return;
      eventq.next = eqp->next;  /* Remove from event queue */
      if ((d = eqp->dp) == NULL)  /* No distributions queued */
         free_event(eqp);
      else {
         do {
            compute_rate(d);
            d = d->nrate;
            } while (d != NULL);
         eqp->t += eqp->interval;
         schedule_event(eqp);      
         }
      }
   }

void link_distrib_to_event(struct distribution *d)
{
   struct event *eqp;
   struct distribution *dq;
   for (eqp = eventq.next; eqp != NULL; eqp = eqp->next) { 
      if (eqp->interval == d->Parameter1) break;
      }
   if (eqp == NULL) {  /* Make new event for this interval */
      eqp = get_event();
      eqp->dp = NULL;
      eqp->interval = d->Parameter1;
      eqp->t = uptime_s() + eqp->interval;
      schedule_event(eqp);
      }
   d->nrate = eqp->dp;  /* Add distrib to head of event chain */
   eqp->dp = d;
   }

#if DIST_DBG
void print_dist_chain(struct event *ep)
{
   struct distribution *dq;
   for (dq = ep->dp; dq != NULL; dq = dq->nrate)
      printf("%d@%lu ", dq->fp->fk.PeerAddrType, dq);
   printf("\n");
   }
#endif

void unlink_distrib_from_event(struct distribution *d)
{
   struct event *eqp;
   struct distribution *dq, *lq;
   for (eqp = eventq.next; eqp != NULL; eqp = eqp->next) { 
      if (eqp->interval == d->Parameter1) break;
      }
#if DIST_DBG
   printf("unlink_distrib(%lu): ", d);  /* $$$ */
   print_dist_chain(eqp);
#endif
   if (eqp == NULL) {
      scpos(0,scr_lrow);
      log_msg(LOG_ERR, 0, 
         "No event for %ds rate distribution <<<", d->Parameter1);
      return;
      }
   for (lq = NULL, dq = eqp->dp; dq != d; lq = dq, dq = dq->nrate) {
      if (dq == NULL) {
         scpos(0,scr_lrow);
         log_msg(LOG_ERR, 0, "Couldn't find rate distribution <<<");
         return;
         }
      }
   if (lq == NULL) eqp->dp = d->nrate;
   else lq->nrate = d->nrate;
#if DIST_DBG
   printf("   ");  /* $$$ */
   print_dist_chain(eqp);
#endif
   }
#endif  /* NEW_ATR */

#if TCP_ATR
static int complained_about_sfd_blocks = 0;
struct subflow_data *get_sf_data(void)  /* Get unused sf_data */
{
   struct subflow_data *sh;
   if (free_sfd_blocks != NULL) {
      sh = free_sfd_blocks;  free_sfd_blocks = free_sfd_blocks->next;
      --n_free_sfd_blocks;
      memset(sh, 0, sizeof(struct subflow_data));
      return sh;
      }
   else {
      if (!complained_about_sfd_blocks) {
         log_msg(LOG_WARNING, 0, "Not enough TCP flow space <<<\n");
         complained_about_sfd_blocks = 1;
         }
      return NULL;
      }
   }

void free_sf_data(struct subflow_data *sh)  /* Return sf_data to free list */
{
   sh->next = free_sfd_blocks;  free_sfd_blocks = sh;
   ++n_free_sfd_blocks;
   complained_about_sfd_blocks = 0;
   }

static int complained_about_subflows = 0;
struct subflow *get_subflow(void)
{
   struct subflow *sf;
   if (free_subflows != NULL) {
      sf = free_subflows;  free_subflows = free_subflows->next_sf;
      --n_free_subflows;
      return sf;
      }
   else {
      if (!complained_about_subflows) {
         log_msg(LOG_WARNING, 0, "Not enough TCP stream space <<<\n");
         complained_about_subflows = 1;
         }
      return NULL;
      }
   }

void free_subflow(struct subflow *sf)
{
   sf->next_sf = free_subflows;  free_subflows = sf;
   ++n_free_subflows;
   complained_about_subflows = 0;
   }
#endif  /* TCP_ATR */

#ifdef DEBUG
FILE *logfl;
int logf_nbr = 1;

void open_log()
{
   char fn[30];
   sprintf(fn,"flows.%03d",logf_nbr++);
   logfl = fopen(fn, "w");
   }

void paddr(Bit8 huge *a)
{
   fprintf(logfl,"%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
   }

void daddr(Bit8 huge *a)
{
   int j;
   printf("%u",a[0]);
/*   for (j = 1; j != RULE_ADDR_LEN; ++j) printf(".%u", a[j]); */
   for (j = 1; j != 4; ++j) printf(".%u", a[j]);
   }

void dport(Bit8 huge *a)
{
   printf("%u",a[0]  << 8 | a[1]);
   }

void fdaddr(FILE *f, Bit8 huge *a)
{
   fprintf(f, "%u.%u.%u.%u",a[0],a[1],a[2],a[3]);
   }

void fdport(FILE *f, Bit8 huge *a)
{
   fprintf(f, "%u",a[0]  << 8 | a[1]);
   }

void pkey(char *msg, struct flow huge *f)
{
   if (!logfl) open_log();
   fprintf(logfl,"%s: %lu  %u\n   {", msg,f, f->fk.PeerAddrType);
   paddr(f->fk.Low.PeerAddress); fprintf(logfl," & ");
   paddr(Masks[f->fk.Low.PeerMaskVal].Val); fprintf(logfl,"} -> {");
   paddr(f->fk.High.PeerAddress); fprintf(logfl," & ");
   paddr(Masks[f->fk.High.PeerMaskVal].Val); fprintf(logfl,"}\n");
   }
#endif

#define DRT_SIZE  3  /* Default rule set: 3 rules */
#ifdef OLD_CT
#define DCT_SIZE  1  /*   1 count */
#endif
struct rule default_rule_table[DRT_SIZE] = {
/* Selector,        Mask, Value,       Action, JumpIndex */
   {FTLOWPEERTYPE,
# if CLNS
      1,  AT_DUMMY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# elif V6 || FULL_IPX
      1,  AT_DUMMY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# else  /* v4 */
      1,  AT_DUMMY,0,0,0,0,0,
# endif
      RA_IGNORE, 0}, /* SourcePeerType & 255 = dummy: Ignore, 0; # 1 */
   {RS_NULL,

# if CLNS
      0,  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# elif V6 || FULL_IPX
      0,  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# else  /* v4 */
      0,  0,0,0,0,0,0,
# endif
      RA_GOTOACT, 3},  /* Null & 0 = 0: GotoAct, Next; # 2 */
   {FTLOWPEERTYPE,

# if CLNS
      1,  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# elif V6 || FULL_IPX
      1,  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
# else  /* v4 */
      1,  0,0,0,0,0,0,
# endif
      RA_COUNTPKT, 1} };  /* SourcePeerType & 255 = 0: CountPkt, 1; # 3 */

Bit32 p2primes[] = {  /* Primes just below powers of two */
   7, 13, 31, 61, 127, 251, 509, 1021,      /*  0 .. 7 */
   2039, 4093, 8191, 16381, 32749, 65521,   /*  8 .. 13 */
   131071, 262139, 524287, 1048573};        /* 14 .. 17 */

int primes[22] = {  /* Two for Trans bytes + enough for NSAP bytes */
   211, 251,
   113, 139, 173, 227, 179, 193, 181, 131,  67,  73, 
    83,  97, 103, 127, 149,  23,  53,  89,  59,  47};

int optimise_rule_table(int rs)
{
   struct rtinf_rec *rip = &ri[rs-1];
   Bit16 crs, j, k, g, h, x, l,
      nht;  /* Nbr of rule_hash_tbls */
   Bit32 ths,  /* Sum of rule_hash_tbl sizes */
      rhss;  /* Total rule_hash sequence size */
   struct rule huge *crt, huge *rp, huge *hp;
   Bit16 huge *rs_rule_hash;
   struct rule_hash_tbl huge *c_rh;
   Bit16 hash;
   char msg[60];

   crt = rip->ri_rule_table;
   crs = rip->ri_Size;
#ifdef H_TEST
   scpos(0,scr_lrow);  printf("\nRT: crs=%u, RULE_ADDR_LEN=%d\n",
      crs,RULE_ADDR_LEN);
#endif
   hp = &crt[0];  hp->hash_tbl_index = 1;
   for (j = 1;  j != crs;  ++j) {  /* Mark match break points */
      rp = &crt[j];
      rp->hash_tbl_index = rp->RuleSelector != hp->RuleSelector ||
         rp->RuleMaskVal != hp->RuleMaskVal;
      hp = rp;
      }
   for (j = 0;  j != crs;  ++j) {  /* Mark goto break points */
      rp = &crt[j];  rp->hash_link = 0;
      k = rp->RuleAction;
      if (k == RA_PUSHTO  /* OK to gotoact in middle of compare group */
            || k == RA_PUSHPKTTO || k == RA_GOTO
            || k == RA_GOSUB     || k == RA_ASSIGN
            || k == RA_POPTO
            ) {
         hp = &crt[rp->RuleJumpIndex-1];
         hp->hash_tbl_index = 2;
         }
      }
#ifdef H_TEST
   scpos(0,scr_lrow);  printf("rule  index  link  MARKED\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      printf("%3d %5d %5d  %d  ",
         j,rp->hash_tbl_index,rp->hash_link, rp->RuleSelector);
      daddr((Bit8 huge *)Masks[rp->RuleMaskVal].Val);
      printf("\n");
      }
#endif
   nht = ths = 0;
   hp = &crt[0];  k = 1;
   for (j = 1;  j != crs;  ++j) {  /* Find compare groups */
      rp = &crt[j];
      if ((g = rp->hash_tbl_index) == 0) ++k;  /* Same group */
      if (g != 0 || j == crs-1) {  /* End of a group */
         if (k >= MNCGRPSZ && hp->RuleSelector != RS_NULL) {
            for (h = 0; h != sizeof(p2primes)/sizeof(Bit32)-1 &&
               p2primes[h] <= (Bit32)k; ++h) ;  /* h = next prime above k */
            if (p2primes[h] > MXRHTSZ) h = MXRHTSZ;
            else h = (Bit16)p2primes[h];
            hp->hash_tbl_index = k;  /* Remember group and */
            hp->hash_link = h;       /*   hash table sizes */
            ++nht;  ths += h;
            }
         else hp->hash_tbl_index = 0;
         hp = rp;  k = 1;
         }
      rp->hash_tbl_index = 0;  /* Last row can't start a group */
      }
   rhss = 1 + nht*2 + ths;  /* rule_hash set size (Bit16s) */
   if (rhss > 0xFFFF) {
      log_msg(LOG_ERR, 0, msg,"Ruleset '%s': rhss = %ul (> 65535)", 
         rip->ri_Name, rhss);
      return 0;  /* Fail */
      }
#ifdef H_TEST
   printf("rule  index  link  GROUPED\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      printf("%3d %5d %5d\n", j, rp->hash_tbl_index,rp->hash_link);
      }
#endif

   if (!alloc_rulespace(rs, 0, rhss))  /* Alloc rule hash space */
      return 0;  /* No room for rule hash table! */

   rs_rule_hash = rip->ri_rule_hash;
   rs_rule_hash[0] = 0;  k = 1;  rp = &crt[j = 0];
   for ( ; nht != 0; --nht, k += 2+h) {
      c_rh = (struct rule_hash_tbl huge *)&rs_rule_hash[k];
      while (rp->hash_tbl_index == 0) {
         ++rp; ++j;
         }
      g = rp->hash_tbl_index;  /* Group size */
      h = rp->hash_link;  /* Hash table size */
      rp->hash_tbl_index = k;  /* Index of hash_tbl in rule_hash */
      rp->hash_link = 0;
      c_rh->group_last = j+g;
      c_rh->hashmod = h;
      for (x = 0; x != h; ++x) c_rh->hash_ent[x] = 0;
      for ( ; g != 0; --g) {
         for (hash = x = 0; x != RULE_ADDR_LEN; ++x)
            hash += rp->RuleMatchedValue[x]*primes[2+x];
         hash %= h;
         if ((l = c_rh->hash_ent[hash]) == 0)
            c_rh->hash_ent[hash] = j+1;  /* First rule with this hash */
         else {
            do {
               hp = &crt[l-1];  l = hp->hash_link;
               } while (l != 0);
            hp->hash_link = j+1;  /* Add to end of hash chain */
            }
         ++rp; ++j;
         }
      }
#ifdef H_TEST
   printf("Rule Hash Tables ..\n\nrule index  n_rules n_hash n_zero\n\n");
   for (j = 0;  j != crs;  ++j) {
      rp = &crt[j];
      if ((k = rp->hash_tbl_index) == 0) continue;
      c_rh = (struct rule_hash_tbl huge *)&rs_rule_hash[k];
      h = c_rh->hashmod+1;
      for (g = 0, hash = 0; hash != h; ++hash)
         if (c_rh->hash_ent[hash] == 0) ++g;
      printf("%5d%5d  %5d%5d%5d\n     ",
         j+1, k, c_rh->group_last-j+1, h, g);
      printf("\n");
      }
#endif
   return 1;
   }

void empty_hash_chain(int rs)  /* Unlink flows for ruleset rs */
{
   Bit32 k;
   struct flow huge * huge *sf;
   struct flow huge *fp, huge *lfp;
   struct flow huge *chp, huge *ctp;

   /* Could walk the rs chain,rather than linear searching the whole
      whole table!  But we don't shut down rule sets very often! */

   for (k = 0; k != fthashmod; ++k) {
      sf = &flow_tbl_htp->hash_ent[k];  /* Address of hash table entry */
      if ((fp = *sf) != NULL) {
         chp = ctp = NULL;  /* New chain for remaining active flows */
         do {
            lfp = fp;  fp = lfp->ht_next;
            if (lfp->FlowRuleSet == rs)
               mark_orphan(lfp);
            else {
               if (chp == NULL) chp = ctp = lfp;
               else {
                  ctp->ht_next = lfp;
                  ctp = lfp;
                  }
               }
            } while (fp != (struct flow huge *)sf);
         if (chp != NULL) {
            ctp->ht_next = fp;
            *sf = chp;
            }
         else {  /* No remaining active flows */
            *sf = NULL;
            --n_hash_ents;
            }
         }
      }
   }

#if PME_TRACE
#define MXTRACE      250
#define MXTRACE_MSGS   5  /* Per rule set */

static Bit16 ruletrace[MXTRACE], rtx;
static int tr_msgs_left[MXRTBLS];
#endif

void open_rule_set(int n, int open_rs)
{
   int k, p;
   Bit16 rsx;
   struct rule huge *rp;
   if (n == 0) return;  /* 0 = no rule set to run */
   if (ri[n-1].ri_Status != RS_ACTIVE) return;

   if (open_rs) {
      if (++ri[n-1].ri_running != 1)  /* Already running */
         return;
#if PME_TRACE
      else tr_msgs_left[n-1] = MXTRACE_MSGS;
#endif
      }
   else {  /* Close ruleset */
      if (--ri[n-1].ri_running != 0)  /* Still running */
         return;
      empty_hash_chain(n);
      }

   c_rt = ri[n-1].ri_rule_table;  c_rsz = ri[n-1].ri_Size;
   for (rsx = 0; rsx != c_rsz; ++rsx) {
      rp = &c_rt[rsx];
      if (rp->RuleSelector == FTLOWPEERTYPE ||
            rp->RuleSelector == FTHIPEERTYPE) {
         p = rp->RuleMatchedValue[0];
         if (rp->RuleAction == RA_PUSHPKTTO ||  /* All peer types */
               rp->RuleAction == RA_PUSHPKTTOACT ||
               rp->RuleAction == RA_COUNTPKT)
            for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k) {
               if (open_rs) ++proto_reqd[k];  
               else --proto_reqd[k];
               }
         else if (p <= MX_PROTOCOLS & *proto_names[p] != '\0') {
            if (open_rs) ++proto_reqd[p];  /* Specific peer type */
            else --proto_reqd[p];
            }
         }
      if ((rp->RuleSelector == FTLOWADJACENTADDRESS ||
            rp->RuleSelector == FTHIADJACENTADDRESS) ||
            ((rp->RuleAction == RA_ASSIGN ||
            rp->RuleAction == RA_ASSIGNACT) &&
            (rp->RuleMatchedValue[0] == FTLOWADJACENTADDRESS ||
            rp->RuleMatchedValue[0] == FTHIADJACENTADDRESS)) )
         if (open_rs) ++adj_reqd;
         else --adj_reqd;
#if NF_OCX_BGP
      if (rp->RuleSelector == FTLOWROUTEASN ||
            rp->RuleMatchedValue[0] == FTHIROUTEASN)
         if (open_rs) ++asn_lookup;
         else --asn_lookup;
#endif
      }
   if (adj_reqd) {
      for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k)
         if (proto_reqd[k]) return;  /* At least one protocol specified */
      for (k = 0; k != sizeof(proto_reqd)/sizeof(int); ++k) {
         if (open_rs) ++proto_reqd[k];  /* All peer types */
         else --proto_reqd[k];
         }
      }
   }

#ifdef GC_DEBUG
int dup_flow(struct flow huge *f, char *msg)
{
   Bit32 x;
   for (x = 253; x <= mxflows; ++x)
      if (flow_ix[x-1] != NULL) {
         scpos(0,scr_lrow);
         printf("   flow_ix[%d]=%lu !!\n", x,flow_ix[x-1]);
         }
   for (x = 1; x <= mxflows; ++x)
      if (flow_ix[x-1] == f) {
         scpos(0,scr_lrow);
         printf("Flow %lu is flow_ix[%d], msg=%s\n", f,x, msg);
         exit(0);
         }
   return 0;
   }
#endif

#if RS_CHAINS
Bit32 rsc_n_searches,
   rsc_mx_range, rsc_av_range,  /* Flow index range searched */
   rsc_mx_tests, rsc_av_tests;  /* Nbr of TimeFilter tests */

# if RS_CH_DEBUG
void check_rs_chain(int rs, char *msg)
{
   Bit32 fx;
   int count;
   for (count = 0, fx = ri[rs-1].ri_flow_chain;
         fx != 0;  fx = fa[fx-1].rs_next) {
      if (++count > ri[rs-1].ri_FlowRecords) break;
      }
   if (count != ri[rs-1].ri_FlowRecords)
      printf("check_rs_chain(%d, %s): count=%d, FlowRecords=%d\n",
         rs,msg, count, ri[rs-1].ri_FlowRecords);
   }
# endif

void mark_idle(Bit32 fx)
{
   struct flow huge *fp = &fa[fx-1];
   fp->FlowRuleSet = 0;   /* Mark idle */
   if (free_flow_tail == 0) free_flow_head = fx;
   else fa[free_flow_tail-1].rs_next = fx;  /* Add to tail of free chain */
   free_flow_tail = fx;
   fp->rs_next = 0;  /* End marker for rs chain */
   }

Bit32 alloc_flow_nbr(void)  /* Allocate nbr for a new a flow */
{
   static int already_complained;
   Bit32 fx;

   if (free_flow_head == 0) {  /* No idle flows: force garbage collection */
      if (garbage_collect(0))
         already_complained = 0;
      }
   if (free_flow_head != 0) {
      fx = free_flow_head;
      free_flow_head = fa[fx-1].rs_next;  /* Take from head of free chain */
      if (free_flow_head == 0) free_flow_tail = 0;
      return fx;
      }
   if (already_complained) return 0;
   more_garbage();  /* Increase garbage collecting rate */
# if defined(DOS)
   scpos(0,scr_lrow);  printf("Recovering flows too slow !!!");
# else
   log_msg(LOG_INFO, 0, "Recovering flows too slow !!!\n");
# endif
   already_complained = 1;
   return 0;
   }
#endif

Bit32 number_flow()  /* Allocate a flow_nbr */
{
   static int already_complained;
   Bit32 start_ix = empty_ix;

   do {
      if (++empty_ix > mxflows)  /* empty_ix goes from 1 to MXFLOWS */
         empty_ix = 1;  /* Flows go from 1 to mxflows too */
      if (flow_idle(empty_ix)) {
#  ifdef GC_DEBUG
         printf("nbr_flow(%lu) -> %u\n", fp,empty_ix);
#  endif
         return empty_ix;  /* 1-org */
         }
      } while (empty_ix != start_ix);

   if (garbage_collect(0)) {  /* No unused flows: force garbage collection */
      already_complained = 0;
      return empty_ix = gcf_ix;  /* 1-org */
      }
   if (already_complained) return 0;
   more_garbage();  /* Increase garbage collecting rate */
#if defined(DOS)
   scpos(0,scr_lrow);  printf("Recovering flows too slow !!!");
#else
   printf("Recovering flows too slow !!!\n");
#endif
   already_complained = 1;
   return 0;
   }

struct flow huge *find_flow(Bit32 x)  /* Find flow with flow_nbr x */
{
   if (x > mxflows || x < 1) {
      log_msg(LOG_ERR, 0, "Invalid flow number (%d)", x);
quack(86); /* @@ */
      return NULL;
      }
#if FLOW_INDEX
   return flow_ix[x-1];
#else
   return &fa[x-1];
#endif
   }

void more_garbage()
{
   int max_gc_f;
   gc_interval >>= 1;  /* Decrease garbage collection interval */
   if (gc_interval < 2) {
      gc_interval = 2;
      max_gc_f = (mxflows)/7;
      gc_f <<= 1;  /* Increase nbr of flows checked each collection */
      if (gc_f > max_gc_f) gc_f = max_gc_f;
      }
   }

int garbage_collect(int incremental)
{
   int f,e;
   Bit32 s_gcf_ix;
   Bit32 GCtime;
   struct rule huge *rp;
   struct flow huge *fp, huge *np, huge *tp, huge *ltp;
#ifdef GC_DEBUG
   int k;
#endif
#if RS_CHAINS
   Bit32 rs_first_flow;
   struct flow huge *pf;
   int rs;
#endif

   e = gc_f;  f = gc_f/4;
#ifdef GC_TEST1
   scpos(0,scr_lrow);
   printf("GC: gcf_ix=%u, e=%d, f=%d\n", gcf_ix,e,f);
#endif
   s_gcf_ix = gcf_ix;
   do {
      if (++gcf_ix > mxflows)  /* gcf_ix goes from 1 to MXFLOWS */
         gcf_ix = 1;
      if (flow_idle(gcf_ix)) {  /* Unused flow index */
         if (incremental && --e == 0) return 1;
         continue;
         }
#if FLOW_INDEX
      fp = flow_ix[gcf_ix-1];
#else
      fp = &fa[gcf_ix-1];
#endif
      GCtime = ri[fp->FlowRuleSet-1].ri_gc_time;
#ifdef GC_TEST1
      printf("   GCtime=%lu, FlowRuleSet=%u, gcf_ix=%lu\n",
         GCtime,fp->FlowRuleSet,gcf_ix);
#endif
      if (fp->ntm_LastTime <= GCtime) {  /* Inactive, recover it */
#ifdef GC_TEST1
         printf("   f=%d, gcf_ix=%d, ntm_LastTime=%lu\n",
            f,gcf_ix,fp->ntm_LastTime);
#endif
         if (!flow_orphan(fp)) {  /* Remove from hash chain */
#ifdef GC_DEBUG
            scpos(0,scr_lrow);
            printf("GC: set=%u, ix=%u, fp=%lu, np=%lu\n",
               fp->FlowRuleSet,gcf_ix, fp,fp->ht_next);
            k = 0;
#endif
            np = fp->ht_next;
            tp = fp;  do {
               ltp = tp;
#ifdef GC_DEBUG
printf("   k=%d, ltp=%lu, tp=%lu\n", ++k, ltp,ltp->ht_next);
if (tp->ht_next == NULL || k > 10) exit(0);  /* Infinite loop - stop */
#endif
               } while ((tp = tp->ht_next) != fp);
            if (ltp == np) {  /* Single flow in hash chain */    
               mark_orphan(ltp);
               --n_hash_ents;
	       }
            else ltp->ht_next = np;
            mark_orphan(fp);  /* Added 15 May 99, Nevil */
            }

         --ri[fp->FlowRuleSet-1].ri_FlowRecords;
#if RS_CHAINS
         rs = fp->FlowRuleSet;
         rs_first_flow = ri[rs-1].ri_flow_chain;
         if (gcf_ix == rs_first_flow) {  /* First flow of chain */
            ri[rs-1].ri_flow_chain = fp->rs_next;
            }
         else {  /* Find previous link in ruleset chain */
            pf = fp;  do {
               --pf;
               } while (pf->FlowRuleSet != rs);
            pf->rs_next = fp->rs_next;
            }
# if RS_CH_DEBUG
         check_rs_chain(rs, "GC");
# endif
#endif
         mark_idle(gcf_ix);

#ifdef GC_DEBUG
         printf("\ngarbage_collect ..\n");
#endif
         free_flow(fp);
         --nflows;  ++FlowsRecovered;
         if (!incremental) return 1;  /* Forced collection succeeded,
                                         flow index gcf_ix is now free */
#ifdef GC_TEST
         scpos(0,scr_lrow);
         printf("Flow %d recovered\n", gcf_ix);
#endif
         }
      else {  /* Not inactive */
#if TCP_ATR
         if (fp->sfdata != NULL)
            check_tcp_idle_subflows(fp, 1);
#endif  /* TCP_ATR */
         }
      if (incremental && --f == 0) return 1;
      } while (gcf_ix != s_gcf_ix);
   return 0;  /* Failed to recover any flows */
   }

#define MIGNORE 2  /* match() result */
#define MSUC    1
#define MFAIL   0
#define MFWD    1  /* count() parameter */
#define MREV    0

struct search_result {
   struct flow huge * huge *sf;  /* Address of hash table entry */
   struct flow huge *lf;  /* Last flow in hash chain */
   };

int match(int forward, struct pkt huge *bp,
      struct pkt_key huge *from, struct pkt_key huge *to);
   void build_search_key(struct pkt huge *bp,
      struct pkt_key huge *Low, struct pkt_key huge *High);
struct flow huge *current(struct search_result*s_r,
   struct key huge *Low, struct key huge *High);
struct flow huge *create(struct search_result *s_r,
   int flow_order, struct pkt huge *bp);
void count(struct flow huge *fp,
   int flow_order, struct pkt huge *bp);

struct flow_key search_key;  /* For searching flow table */
int pal;
Bit32 search_hash;

struct search_result sd_pointers, ds_pointers;

Bit16 p_stack[RSTKSIZ];  /* Pattern stack */
   /* Was int p_stack.  22 Apr 98 */
int a_stack[RSTKSIZ];  /* Attribute values for pattern stack */
int p_s_depth;

Bit8 user_vbls[N_VBLS];  /* User variables */
Bit16 env_stack[ESTKSIZ];  /* Environment stack */
int e_s_tos;

#ifdef GC_DEBUG
void print_chain(struct hash_tbl huge *t, Bit16 hash, char *msg)
{
   struct flow huge * huge *sf;
   struct flow huge *f, huge *lf;
   sf = &t->hash_ent[hash];  /* Address of hash table entry */
   scpos(0,scr_lrow);  printf("Chain: t=%lu, hash=%u, sf=%lu, msg=%s\n",
      t,hash,sf, msg);
   lf = f = sf[0];
   do {
      printf("   f=%lu\n", f);
      if ((lf = f) == NULL) return;
      } while ((f = lf->ht_next) != (struct flow huge *)sf);
   }
#endif

void build_search_key(
   struct pkt huge *bp,
   struct pkt_key huge *Low,  /* Target keys */
   struct pkt_key huge *High
   )
{
   struct rule huge *rp;
   int j, k, push_rule;
   Bit8 *Maskp;

   search_hash = c_rtx*139;
      /* Include ruleset ix in search key.  Nevil, 11 Sep 98 */
   memset(&search_key, 0, sizeof(search_key));  /* Start with a null key */
   pal = PEER_ADDR_LEN;  /* Could be any type, start by using max length */
#ifdef TESTING
   if (!logfl) open_log();
   fprintf(logfl,"p_s_depth=%d\n", p_s_depth);
#endif
   for (j = 0; j != p_s_depth; ++j) {
      rp = &c_rt[p_stack[j]-1];
      Maskp = Masks[rp->RuleMaskVal].Val;
#ifdef TESTING
      fprintf(logfl,"   j=%d, attrib=%d\n", j,a_stack[j]);
#endif
      push_rule = (rp->RuleAction != RA_PUSHPKTTO) &&
         (rp->RuleAction != RA_PUSHPKTTOACT) &&
         (rp->RuleAction != RA_COUNTPKT);
      switch (a_stack[j]) {
      case FTLOWPEERADDRESS:
         search_key.Low.PeerMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != pal; ++k)
            search_hash += (search_key.Low.PeerAddress[k] =
               rp->RuleMatchedValue[k])*primes[2+k];
         else for (k = 0; k != pal; ++k)
            search_hash += (search_key.Low.PeerAddress[k] =
               Low->PeerAddress[k] & Maskp[k])*primes[2+k];
         break;
      case FTHIPEERADDRESS:
         search_key.High.PeerMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != pal; ++k)
            search_hash += (search_key.High.PeerAddress[k] =
               rp->RuleMatchedValue[k])*primes[2+k];
         else for (k = 0; k != pal; ++k)
           search_hash += (search_key.High.PeerAddress[k] =
              High->PeerAddress[k] & Maskp[k])*primes[2+k];
         break;
      case FTLOWPEERTYPE:
      case FTHIPEERTYPE:
         if (push_rule) search_hash +=
            (search_key.PeerAddrType = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.PeerAddrType = bp->PeerAddrType);
         pal = addr_len[search_key.PeerAddrType];  /* Now we know peer type */
         break;
      case FTLOWTRANSADDRESS:
         search_key.Low.TransMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.Low.TransAddress[k] =
               rp->RuleMatchedValue[k])*primes[k];
         else for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.Low.TransAddress[k] =
               Low->TransAddress[k] & Maskp[k])*primes[k];
         break;
      case FTHITRANSADDRESS:
         search_key.High.TransMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.High.TransAddress[k] =
               rp->RuleMatchedValue[k])*primes[k];
         else for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.High.TransAddress[k] =
               High->TransAddress[k] & Maskp[k])*primes[k];
         break;
      case FTLOWTRANSTYPE:
      case FTHITRANSTYPE:
         if (push_rule) search_hash +=
            (search_key.TransAddrType = rp->RuleMatchedValue[0])*241;
         else search_hash +=
            (search_key.TransAddrType = bp->TransAddrType)*241;
         break;
      case FTLOWADJACENTADDRESS:
         search_key.Low.AdjMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != MAC_ADDR_LEN; ++k)
            search_hash += (search_key.Low.AdjAddress[k] =
               rp->RuleMatchedValue[k]);
         else for (k = 0; k != MAC_ADDR_LEN; ++k)
            search_hash += (search_key.Low.AdjAddress[k] =
               Low->AdjAddress[k] & Maskp[k]);
         break;
      case FTHIADJACENTADDRESS:
         search_key.High.AdjMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != MAC_ADDR_LEN; ++k)
            search_hash += (search_key.High.AdjAddress[k] =
               rp->RuleMatchedValue[k]);
         else for (k = 0; k != MAC_ADDR_LEN; ++k)
            search_hash += (search_key.High.AdjAddress[k] =
               High->AdjAddress[k] & Maskp[k]);
         break;
      case FTSOURCECLASS:
         search_hash += (search_key.Low.Class = rp->RuleMatchedValue[0]);
         break;
      case FTDESTCLASS:
         search_hash += (search_key.High.Class = rp->RuleMatchedValue[0]);
         break;
      case FTFLOWCLASS:
         search_hash += (search_key.FlowClass = rp->RuleMatchedValue[0]);
         break;
      case FTSOURCEKIND:
         search_hash += (search_key.Low.Kind = rp->RuleMatchedValue[0]);
         break;
      case FTDESTKIND:
         search_hash += (search_key.High.Kind = rp->RuleMatchedValue[0]);
         break;
      case FTFLOWKIND:
         search_hash += (search_key.FlowKind = rp->RuleMatchedValue[0]);
         break;
      case FTLOWINTERFACE:
         if (push_rule) search_hash +=
            (search_key.Low.Interface = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.Low.Interface = Low->Interface);
         break;
      case FTHIINTERFACE:
         if (push_rule) search_hash +=
            (search_key.High.Interface = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.High.Interface = High->Interface);
         break;
      case FTLOWADJACENTTYPE:
         if (push_rule) search_hash +=
            (search_key.Low.AdjAddrType = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.Low.AdjAddrType = Low->AdjType);
         break;
      case FTHIADJACENTTYPE:
         if (push_rule) search_hash +=
            (search_key.High.AdjAddrType = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.High.AdjAddrType = High->AdjType);
         break;
      case FTDSCODEPOINT:
         if (push_rule) search_hash +=
            (search_key.DSCodePoint = rp->RuleMatchedValue[0])*151;
         else search_hash +=
            (search_key.DSCodePoint = bp->DSCodePoint)*151;
         break;
#if NF_ASN_ATT
      case FTLOWROUTEASN:
         search_key.Low.ASNMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.Low.RouteASN[k] =
               rp->RuleMatchedValue[k]);
         else for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.Low.RouteASN[k] =
               Low->nf_ASN[k] & Maskp[k]);
         break;
      case FTHIROUTEASN:
         search_key.High.ASNMaskVal = rp->RuleMaskVal;
         if (push_rule) for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.High.RouteASN[k] =
               rp->RuleMatchedValue[k]);
         else for (k = 0; k != TRANS_ADDR_LEN; ++k)
            search_hash += (search_key.High.RouteASN[k] =
               High->nf_ASN[k] & Maskp[k]);
         break;
      case FTLOWROUTEPREFIX:
         if (push_rule) search_hash +=
            (search_key.Low.RoutePrefix = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.Low.RoutePrefix = Low->nf_mask);
         break;
      case FTHIROUTEPREFIX:
         if (push_rule) search_hash +=
            (search_key.High.RoutePrefix = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.High.RoutePrefix = High->nf_mask);
         break;
#endif
#if NF_OTHER_ATT
      case FTMETERID:
         if (push_rule) search_hash +=
            (search_key.MeterId = rp->RuleMatchedValue[0]);
         else search_hash +=
            (search_key.MeterId = bp->ntm_interface);
         break;
#endif
         }
      }
#ifdef MATCH_TRACE
   printf("\nType,Class,Kind: %d  %d %d %d  %d %d %d  hash=%d\n",
      search_key.PeerAddrType,
      search_key.Low.Class,search_key.High.Class,search_key.FlowClass,
      search_key.Low.Kind,search_key.High.Kind,search_key.FlowKind,
      search_hash);
#endif
   }

#ifdef STACK_CHECK
void dump_pat_stack(char *reason)
{
   int j;
   scpos(0,24);
   printf("match %s: stk=", reason);
   for (j = 0; j != RSTKSIZ; ++j)
      printf("%u+%u,", a_stack[j],p_stack[j]);
   printf("\n   LowPeer=");  daddr(search_key.Low.PeerAddress);
   printf(", HighPeer=");  daddr(search_key.High.PeerAddress);
   printf("\n");
   }
#endif

#if PME_TRACE
void log_PME_TRACE(void)
{
   int ptx;
   for (ptx = 0; ptx <= rtx-1; ptx+=10)
      log_msg(LOG_INFO, 0, "   %d, %d, %d, %d, %d, %d, %d, %d, %d, %d,",
         ruletrace[ptx], ruletrace[ptx+1], ruletrace[ptx+2],
         ruletrace[ptx+3], ruletrace[ptx+4], ruletrace[ptx+5],
         ruletrace[ptx+6], ruletrace[ptx+7], ruletrace[ptx+8],
         ruletrace[ptx+9]);
   }
#endif

Bit8 masked_value[RULE_ADDR_LEN];

Bit16 mk_hash(Bit8 huge *vp, Bit8 huge *mp, int n)
{
   Bit16 hash;
   int j;
   for (j = hash = 0;  j != n;  ++j)
      hash += (masked_value[j] = vp[j] & mp[j])*primes[2+j];
   return hash;
   }

Bit8  /* Computed attribute values for this packet */
   psc,pdc,pfc, psk,pdk,pfk;

int match(int forward, struct pkt huge *bp,
   struct pkt_key huge *from, struct pkt_key huge *to)
{
   Bit16 rx;
   int j, attrib, match, test_reqd,
      pal, ral;  /* Rule address length */
   struct rule huge *rp;
   Bit16 hash;
   Bit8 *Maskp, mask0, v0;
   struct flow huge *fp;
   struct rule_hash_tbl huge *c_rh;

#if PME_TRACE
   rtx = 0;  /* Initialise match trace */
#endif

   e_s_tos = p_s_depth = 0;  test_reqd = 1;
   psc = pdc = pfc = psk = pdk = pfk = 0;
   pal = PEER_ADDR_LEN;
   for (rx = 1; rx <= c_rsz; ) {  /* Search the rule table */
#if PME_TRACE
      if (rtx != MXTRACE) ruletrace[rtx++] = rx;
      else {
         if (tr_msgs_left[c_rtx-1] != 0) {
            log_msg(LOG_ERR, 0, "PME match loop, ri[%d]: rtx=%d", c_rtx, rtx); 
            log_PME_TRACE();
            --tr_msgs_left[c_rtx-1];
	    }
         quack(123);
         return MFAIL;
         }
#endif
      rp = &c_rt[rx-1];
      Maskp = Masks[rp->RuleMaskVal].Val;
      mask0 = Maskp[0];  /* Byte 0 of mask */
      v0 = rp->RuleMatchedValue[0];  /* Byte 0 of value */
      attrib = rp->RuleSelector;
      if (USER_ATTRIB(attrib))  /* User variable */
         attrib = user_vbls[rp->RuleSelector-FTV1];  /* Dereference it */
      if (test_reqd) {
         if (bp->PeerAddrType != AT_DUMMY) ++n_matches;
         if (rp->hash_tbl_index == 0) {  /* Simple compare */
            switch (attrib) {
            case FTLOWPEERADDRESS:
               for (j = 0;  j != pal;  ++j) {
                  if ((from->PeerAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == pal;
               break;
            case FTHIPEERADDRESS:
               for (j = 0;  j != pal;  ++j) {
                  if ((to->PeerAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == pal;
               break;
            case FTLOWPEERTYPE:
            case FTHIPEERTYPE:
               match = bp->PeerAddrType == v0;
               if (match) pal = addr_len[bp->PeerAddrType];
               break;
            case FTLOWTRANSADDRESS:
               for (j = 0;  j != TRANS_ADDR_LEN;  ++j) {
                  if ((from->TransAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == TRANS_ADDR_LEN;
               break;
            case FTHITRANSADDRESS:
               for (j = 0;  j != TRANS_ADDR_LEN;  ++j) {
                  if ((to->TransAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == TRANS_ADDR_LEN;
               break;
            case FTLOWTRANSTYPE:
            case FTHITRANSTYPE:
               match = bp->TransAddrType == v0;
               break;
            case FTLOWADJACENTADDRESS:
               for (j = 0;  j != MAC_ADDR_LEN;  ++j) {
                  if ((from->AdjAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == MAC_ADDR_LEN;
               break;
            case FTHIADJACENTADDRESS:
               for (j = 0;  j != MAC_ADDR_LEN;  ++j) {
                  if ((to->AdjAddress[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == MAC_ADDR_LEN;
               break;
            case FTFORWARD:  /* Addresses in packet order */
               match = forward == v0;
               break;
            case FTSOURCECLASS:
               match = (psc & mask0) == v0;
               break;
            case FTDESTCLASS:
               match = (pdc & mask0) == v0;
               break;
            case FTFLOWCLASS:
               match = (pfc & mask0) == v0;
               break;
            case FTSOURCEKIND:
               match = (psk & mask0) == v0;
               break;
            case FTDESTKIND:
               match = (pdk & mask0) == v0;
               break;
            case FTFLOWKIND:
               match = (pfk & mask0) == v0;
               break;
            case FTLOWINTERFACE:
               match = from->Interface == v0;  /* Kevin, 23 Nov 98 */
               break;
            case FTHIINTERFACE:
               match = to->Interface == v0;  /* Kevin, 23 Nov 98 */
               break;
            case FTLOWADJACENTTYPE:
               match = from->AdjType == v0;
               break;
            case FTHIADJACENTTYPE:
               match = to->AdjType == v0;
               break;
            case FTDSCODEPOINT:
               match = bp->DSCodePoint == v0;
               break;
#if NF_ASN_ATT
            case FTLOWROUTEASN:
               for (j = 0;  j != TRANS_ADDR_LEN;  ++j) {
                  if ((from->nf_ASN[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == TRANS_ADDR_LEN;
               break;
            case FTHIROUTEASN:
               for (j = 0;  j != TRANS_ADDR_LEN;  ++j) {
                  if ((to->nf_ASN[j] & Maskp[j]) !=
                     rp->RuleMatchedValue[j]) break;
                  }
               match = j == TRANS_ADDR_LEN;
               break;
            case FTLOWROUTEPREFIX:
               match = from->nf_mask == v0;
               break;
            case FTHIROUTEPREFIX:
               match = to->nf_mask == v0;
               break;
#endif
#if NF_OTHER_ATT
            case FTMETERID:
               match = bp->ntm_interface == v0;
               break;
#endif
            case RS_NULL:
            default:
               match = 1;  /* Don't test anything */
            }
         }
      else {  /* Hashed compare */
         match = 0;
         switch (attrib) {
            case FTLOWPEERADDRESS:
               hash = mk_hash(from->PeerAddress,Maskp,ral=pal);
               break;
            case FTHIPEERADDRESS:
               hash = mk_hash(to->PeerAddress,Maskp,ral=pal);
               break;
            case FTLOWPEERTYPE:
            case FTHIPEERTYPE:
               hash = mk_hash(&bp->PeerAddrType,Maskp,ral=1);
               break;
            case FTLOWTRANSADDRESS:
               hash = mk_hash(from->TransAddress,Maskp,
               ral=TRANS_ADDR_LEN);
               break;
            case FTHITRANSADDRESS:
               hash = mk_hash(to->TransAddress,Maskp,
               ral=TRANS_ADDR_LEN);
               break;
            case FTLOWTRANSTYPE:
            case FTHITRANSTYPE:
               hash = mk_hash(&bp->TransAddrType,Maskp,ral=1);
               break;
            case FTLOWADJACENTADDRESS:
               hash = mk_hash(from->AdjAddress,Maskp,ral=MAC_ADDR_LEN);
               break;
            case FTHIADJACENTADDRESS:
               hash = mk_hash(to->AdjAddress,Maskp,ral=MAC_ADDR_LEN);
               break;
            case FTSOURCECLASS:
               hash = mk_hash(&psc,Maskp,ral=1);
               break;
            case FTDESTCLASS:
               hash = mk_hash(&pdc,Maskp,ral=1);
               break;
            case FTFLOWCLASS:
               hash = mk_hash(&pfc,Maskp,ral=1);
               break;
            case FTSOURCEKIND:
               hash = mk_hash(&psk,Maskp,ral=1);
               break;
            case FTDESTKIND:
               hash = mk_hash(&pdk,Maskp,ral=1);
               break;
            case FTFLOWKIND:
               hash = mk_hash(&pfk,Maskp,ral=1);
               break;
            case FTLOWINTERFACE:
               hash = mk_hash(&from->Interface,Maskp,ral=1);
               break;
            case FTHIINTERFACE:
               hash = mk_hash(&to->Interface,Maskp,ral=1);
               break;
            case FTLOWADJACENTTYPE:
               hash = mk_hash(&from->AdjType,Maskp,ral=1);
               break;
            case FTHIADJACENTTYPE:
               hash = mk_hash(&to->AdjType,Maskp,ral=1);
               break;
            case FTDSCODEPOINT:
               hash = mk_hash(&bp->DSCodePoint,Maskp,ral=1);
               break;

#if NF_ASN_ATT
            case FTLOWROUTEASN:
               hash = mk_hash(from->nf_ASN,Maskp,ral=TRANS_ADDR_LEN);
               break;
            case FTHIROUTEASN:
               hash = mk_hash(to->nf_ASN,Maskp,ral=TRANS_ADDR_LEN);
               break;
            case FTLOWROUTEPREFIX:
               hash = mk_hash(&from->nf_mask,Maskp,ral=1);
               break;
            case FTHIROUTEPREFIX:
               hash = mk_hash(&to->nf_mask,Maskp,ral=1);
               break;
#endif
#if NF_OTHER_ATT
            case FTMETERID:
               hash = mk_hash(&bp->ntm_interface,Maskp,ral=1);
               break;
#endif
            case RS_NULL:
            default:
               match = 1;  /* Don't test anything */
               }
            if (!match) {
               c_rh = (struct rule_hash_tbl huge *)&c_rht[rp->hash_tbl_index];
               hash %= c_rh->hashmod;
#ifdef P_TRACE
               ptx = rx;
#endif
               if ((rx = c_rh->hash_ent[hash]) != 0) {
                  do {
                     rp = &c_rt[rx-1];
                     if (qcmp((Bit8 far *)masked_value,
                           rp->RuleMatchedValue, ral) == 0) {
                        match = 1;  break;  /* Matched */
                        }
                     } while ((rx = rp->hash_link) != 0);
                  }
               }
            if (match == 0) rx = c_rh->group_last;  /* Not in group */
#ifdef P_TRACE
            printf("Hash cmp: rx=%d, att=%d, mtch=%d, m_rx=%d, act=%d, ix=%d\n",
               ptx,attrib, match,rx, rp->RuleAction,rp->RuleJumpIndex);
#endif
            }
         }
      else {  /* No test required */
#ifdef P_TRACE
         printf("No test: rx=%d, attrib=%d, action=%d, index=%d\n",
            rx,attrib, rp->RuleAction,rp->RuleJumpIndex);
#endif
         test_reqd = 1;  /* Didn't test for this rule */
         match = 1;
         }
      if (match) {
         switch (rp->RuleAction) {
         case RA_IGNORE:
#if PM_DEBUG
            if (c_rtx == pm_dbg) {
               sprintf(dbg_msg, " f=%d, rx=%d, IGNORE", forward,rx);
               display_msg(0,dbg_msg);
               }
#endif
            return MIGNORE;
         case RA_RETRY:
#if PM_DEBUG
            if (c_rtx == pm_dbg) {
               sprintf(dbg_msg, " f=%d, rx=%d, RETRY", forward,rx);
               display_msg(0,dbg_msg);
               }
#endif
            return MFAIL;
         case RA_COUNT:
         case RA_COUNTPKT:
            if (attrib != 0 || rp->RuleMaskVal != 0) {  /* Non-zero mask */
                  /* Count rule specified an attribute */
               a_stack[p_s_depth] = attrib;  /* Remember last matched rule */
               p_stack[p_s_depth++] = rx;
               }
#ifdef COUNT_TRACE
//            if ((from->PeerAddress[2] == 4 && from->PeerAddress[3] == 33) ||
//                  (to->PeerAddress[2] == 4 && to->PeerAddress[3] == 33))
                  {
               printf("Count: %u+%u",a_stack[0],p_stack[0]);
               for (j = 1; j != p_s_depth; ++j) printf(",%u+%u, ",
                  a_stack[j],p_stack[j]);
               daddr(from->PeerAddress);  printf(":");
               dport(from->TransAddress);
               printf(" %s ", forward ? "->" : "<-");
               daddr(to->PeerAddress);  printf(":");
               dport(to->TransAddress);  printf("\n");
               }
#endif
            build_search_key(bp, from,to);
#if PM_DEBUG
            if (c_rtx == pm_dbg) {
               sprintf(dbg_msg, " f=%d, rx=%d, COUNT", forward,rx);
               display_msg(0,dbg_msg);
               }
#endif
            return MSUC;
         case RA_RETURN:
            --e_s_tos;
#ifdef STACK_CHECK
            if (e_s_tos < 0) {
               scpos(0,scr_lrow);
               printf("gosub oflo: depth=%d, stk=", e_s_tos);
               for (j = 0; j != ESTKSIZ; ++j) printf("%u,", env_stack[j]);
               printf("\n");
               return MFAIL;
               }
#endif
#ifdef P_TRACE
            scpos(0,scr_lrow);
            printf("Return: rx=%d, attrib=%d, gosub@%d, index=%d\n",
               rx,attrib, env_stack[e_s_tos],rp->RuleJumpIndex);
#endif
            rx = env_stack[e_s_tos] + rp->RuleJumpIndex;
            test_reqd = 0;  /* Always return to Action */
            continue;
         case RA_ASSIGN:
         case RA_ASSIGNACT:
            user_vbls[rp->RuleSelector-FTV1]
               = v0;  /* Used to be RuleJumpIndex */
            /* ++rx;  Bug!  Nevil, 6 Mar 97 */
            rx = rp->RuleJumpIndex;
            if (rp->RuleAction == RA_ASSIGNACT) test_reqd = 0;
            continue;
         case RA_GOSUB:
         case RA_GOSUBACT:
            env_stack[e_s_tos++] = rx;  /* Save index of gosub rule */
#ifdef STACK_CHECK
            if (e_s_tos >= ESTKSIZ) {
               scpos(0,scr_lrow);
               printf("gosub oflo: depth=%d, stk=", e_s_tos);
               for (j = 0; j != e_s_tos; ++j) printf("%u,", env_stack[j]);
               printf("\n");
               return MFAIL;
               }
#endif
            rx = rp->RuleJumpIndex;
            if (rp->RuleAction == RA_GOSUBACT) test_reqd = 0;
            continue;
         case RA_GOTO:
         case RA_GOTOACT:
            rx = rp->RuleJumpIndex;
#ifdef TESTING
            scpos(0,scr_lrow);  printf("   GOTO %d\n", rx);
#endif
            if (rp->RuleAction == RA_GOTOACT) test_reqd = 0;
            continue;
         case RA_PUSHTO:
         case RA_PUSHTOACT:
            if (attrib >= FTSOURCECLASS && attrib <= FTFLOWKIND)
                  switch (attrib) {  /* Save 'packet' computed attrib value */
            case FTSOURCECLASS:
               psc = v0;  break;
            case FTDESTCLASS:
               pdc = v0;  break;
            case FTFLOWCLASS:
               pfc = v0;  break;
            case FTSOURCEKIND:
               psk = v0;  break;
            case FTDESTKIND:
               pdk = v0;  break;
            case FTFLOWKIND:
               pfk = v0;  break;
               }
         case RA_PUSHPKTTO:
         case RA_PUSHPKTTOACT:
            a_stack[p_s_depth] = attrib;  /* Remember which rules matched */
            p_stack[p_s_depth++] = rx;
            rx = rp->RuleJumpIndex;
            if (rp->RuleAction == RA_PUSHTOACT ||
                  rp->RuleAction == RA_PUSHPKTTOACT)
                test_reqd = 0;
#ifdef STACK_CHECK
            if (p_s_depth >= RSTKSIZ) {
               dump_pat_stack("oflo");
               return MFAIL;
               }
#endif
            continue;
	 case RA_POPTO:
	 case RA_POPTOACT:
	    --p_s_depth;  /* Forget last matched rule */
	    rx = rp->RuleJumpIndex;
	    if (rp->RuleAction == RA_POPTOACT)
               test_reqd = 0;
#ifdef STACK_CHECK
	    if (p_s_depth < 0) {
               dump_pat_stack("uflo");
	       return MFAIL;
	       }
#endif
	    continue;

         default:  /* Not a valid action! */
            log_msg(LOG_ERR, 0, 
               "Bad action, ruleset %d, rule %d", c_rtx,rx);
#if PME_TRACE
            log_PME_TRACE();
#endif
            return MFAIL;
            }
         }
      else ++rx;
      }
   log_msg(LOG_ERR, 0, "Bad index, ruleset %d, rule %d", c_rtx,rx);
#if PME_TRACE
   log_PME_TRACE();
   quack(456);
#endif
   return MFAIL;  /* Not matched */
   }

struct flow huge *current(
   struct search_result *s_r,
   struct key huge *Low, struct key huge *High)
{
   struct flow huge * huge *sf;
   struct flow huge *f, huge *lf;

   sf = &flow_tbl_htp->hash_ent[search_hash %= fthashmod];
      /* Address of hash table entry */
#ifdef MATCH_TRACE
   printf("  current(): cht=%Fp, sf=%Fp\n", cht,sf);
#endif
   ++n_hash_searches;
   if ((f = sf[0]) == NULL) {
      lf = NULL;  /* Empty hash chain */
      ++n_hash_compares;  /* Make sure compares >= searches! */
      }
   else {
      do {
         ++n_hash_compares;
         for (;;) {  /* Compare whole fields - data has been masked */
            if (qcmp((Bit8 far *)Low,
              (Bit8 far *)&f->fk.Low,
              sizeof(struct key)) != 0) break;
            if (qcmp((Bit8 far *)High,
              (Bit8 far *)&f->fk.High,
              sizeof(struct key)) != 0) break;
            if (qcmp((Bit8 far *)&search_key,
              (Bit8 far *)&f->fk,
              SZFLOWATTRIBS) != 0) break;
            return f;  /* Matched */
            }
         lf = f;
         } while ((f = lf->ht_next) != (struct flow huge *)sf);
      f = NULL;  /* Loop stops with lf -> last flow in chain */
      }
   s_r->sf = sf;  s_r->lf = lf;
      /* Save sf and lf; we'll need them to insert a new flow */
#ifdef MATCH_TRACE
   printf("  current() -> sf=%Fp, lf=%Fp\n", s_r->sf,s_r->lf);
#endif
   return f;
   }

#if NEW_ATR
int make_distrib_list(  /* From pattern stack */
   struct flow huge *f)
{
   struct rule huge *rp;
   int j;
   struct distribution *r, *d, *ld;
   double N,D;

   if (f->distrib_list != NULL) {  /* $$$ */
      scpos(0,scr_lrow);
      log_msg(LOG_ERR, 0, "Non-empty distribution list <<<");
      return 0;
      }
   for (j = 0; j != p_s_depth; ++j) {
      switch (a_stack[j]) {
      case FTTOPACKETSIZE:
      case FTFROMPACKETSIZE:
      case FTTOINTERARRIVALTIME:
      case FTFROMINTERARRIVALTIME:
      case FTTOTURNAROUNDTIME:
      case FTFROMTURNAROUNDTIME:
      case FTTOBITRATE:
      case FTFROMBITRATE:
      case FTTOPDURATE:
      case FTFROMPDURATE:
#if TCP_ATR
      case FTTCPTIME1:
      case FTTCPTIME2:
      case FTTOTCPSIZE:
      case FTFROMTCPSIZE:
      case FTTOTCPRATE1:
      case FTFROMTCPRATE1:
      case FTTOTCPRATE2:
      case FTFROMTCPRATE2:
#endif  /* TCP_ATR */
         rp = &c_rt[p_stack[j]-1];
         if ((d = get_dist()) == NULL) {
            scpos(0,scr_lrow);
            log_msg(LOG_ERR, 0,
               "No mem for %d distribution **", rp->RuleSelector);
            return 0;
            }
         memset(d, 0, sizeof(struct distribution));
         d->selector = rp->RuleSelector;
         memcpy(d->mask_params, Masks[rp->RuleMaskVal].Val, RULE_ADDR_LEN);
         memcpy(d->value_params, rp->RuleMatchedValue, RULE_ADDR_LEN);

         d->Transform = d->mask_params[0];
         d->ScaleFactor = d->mask_params[1];
         d->LowerLimit = (d->mask_params[2] << 8) | d->mask_params[3];
         d->UpperLimit = (d->mask_params[4] << 8) | d->mask_params[5];
         d->Buckets = d->value_params[0];
         if (d->Buckets > MXBUCKETS) {
            scpos(0,scr_lrow);
            log_msg(LOG_ERR, 0,
               "Distribution with > %d buckets **", MXBUCKETS);
            d->Buckets = d->value_params[0] = MXBUCKETS;
	    }
         d->Parameter1 = d->value_params[1];
         d->Parameter2 = (d->value_params[2] << 8) | d->value_params[3];
         d->Parameter3 = (d->value_params[4] << 8) | d->value_params[5];

         N = d->Buckets - 1;
         switch (d->Transform) {
         case DS_LIN:
            D = d->UpperLimit - d->LowerLimit;
            break;
         case DS_LOG:
            D = log10((double)d->UpperLimit/(double)d->LowerLimit);
            break;
            }
         d->M = N/D;

         if (f->distrib_list == NULL)
            f->distrib_list = d;
         else ld->next = d;
         ld = d;

         f->distrib_bits |= 1 << d->selector-FTTOPACKETSIZE;

         switch (a_stack[j]) {
         case FTTOBITRATE:
         case FTFROMBITRATE:
         case FTTOPDURATE:
         case FTFROMPDURATE:
            d->fp = f;
            link_distrib_to_event(d);
            break;
         default:
            d->fp = NULL;
            d->nrate = NULL;
            }
	 }
      }
   return 1;
   }
#endif

#if TCP_ATR
void check_subflow_chains(struct subflow_data huge *sfd, char *msg)
{
   struct subflow **sf;
   struct subflow *sp, *lsp, *nsf;
   int j, pr;
#if 0
   pr = 0;

   sp = sfd->sfc;
   if (pr) printf("sf_chains(): sfd=%lx, sf: %lx",
      sfd, sp);
   while (sp->next_sf != NULL) {
      sp = sp->next_sf;
      if (pr) printf(" -> %lx", sp);
      }
   if (pr) printf("\n");
   for (j = 0; j != SUBFLOW_HASHMOD; ++j) {   
      sf = &sfd->sfp[j];
      if ((sp = *sf) != NULL) {
         if (pr) printf("   j=%d, hc: %lx", j, sp);
         for (;;) {
            lsp = sp;
            if ((sp = lsp->next_hc) == (struct subflow *)sf)
               break;
            if (pr) printf(" -> %0lx", sp);
            }
         if (pr) printf("\n");
         }
      }
#endif
   }

void print_subflow_name(FILE *f, char *msg, 
   struct subflow huge *sp)
{
   fprintf(f, "sf(%s) ", msg);
   fdaddr(f, sp->sfn.SrcPeerAddr);  fprintf(f, ", ");
   fdport(f, sp->sfn.SrcTransAddr);  fprintf(f, " -> ");
   fdaddr(f, sp->sfn.DestPeerAddr);  fprintf(f, ", ");
   fdport(f, sp->sfn.DestTransAddr);
   fprintf(f, ", rs=%d: ", sp->sfn.RuleSet);
   }

void terminate_subflow(
   struct subflow_data *sfd, struct subflow *sp)
{
   struct subflow *np, *tp, *ltp;

/* printf("terminate(): sfd=%0lx, sp=%0lx\n", sfd, sp); */
check_subflow_chains(sfd, NULL);
   np = sp->next_hc;  /* Remove subflow from hash chain */
   tp = sp;  do {
      ltp = tp;
      } while ((tp = tp->next_hc) != sp);
   if (ltp == np)  /* Single flow in hash chain */
      ltp->next_hc = NULL;
   else ltp->next_hc = np;
   }

void terminate_all_subflows(
   struct subflow_data *sfd)
{
   struct subflow *sp, *lsp;
   if ((sp = sfd->sfc) != NULL) {
      do {
         lsp = sp;  sp = sp->next_sf;
         terminate_subflow(sfd, lsp);
         free_subflow(lsp);
         } while (sp != NULL);
   /*   printf("term_all(): sfd=%lx, free_sf=%d\n", 
	sfd, n_free_subflows); @@@ */
      }
   }

int make_tcp_data(  /* From pattern stack */
   struct flow huge *f)
{
   int j,pt, flowIP,flowTCP;

   f->sfdata = NULL;
   flowIP = flowTCP = 0;
   for (j = 0; j != p_s_depth; ++j) {
      switch (a_stack[j]) {
      case FTLOWPEERTYPE:
      case FTHIPEERTYPE:
         pt = c_rt[p_stack[j]-1].RuleMatchedValue[0];
         if (pt == AT_IP4
#if V6
               || pt == AT_IP6
#endif
               )
            flowIP = 1;
         break;
      case FTLOWTRANSTYPE:
      case FTHITRANSTYPE:
         if (c_rt[p_stack[j]-1].RuleMatchedValue[0] == PT_TCP)
            flowTCP = 1;
         break;

      case FTTCPDATA:
	 if (flowIP && flowTCP) {
            f->sfdata = get_sf_data();
            return f->sfdata != NULL;
	    }
         }
      }
   return 0;
   }

int seqcmp(Bit32 seq1, Bit32 seq2)
/* Compare two tcp sequence numbers, allowing for wrap-around */
{
   if (seq1 == seq2) return 0;
   if (seq1 < seq2)
      return seq2-seq1 > 0x0fffffff ? +1 : -1;
   else  /* seq1 > seq2 */
      return seq1-seq2 <= 0x0fffffff ? +1 : -1;
   }

void check_tcp_idle_subflows(struct flow huge *fp, int gc_call)
{
   struct subflow *sp, *lsp, *nsf;
   Bit32 InactTime;
   struct subflow_data *sfd;
#if NEW_ATR
   struct distribution *d;
   Bit32 bytes, b2;
   double delta;
   Bit32 n;
#define LowerSize  Parameter2  /* Bit16 */
#define UpperSize  Parameter3
#endif
#if TRACE_NEG_BLP
   double ToBLF, FromBLF;
   int bad_blp, hdr_printed;
#endif
   InactTime = fp->ntm_LastTime - SUBFLOW_TIMEOUT;
   sfd = fp->sfdata;

#if TRACE_NEG_BLP
   FromBLF = ToBLF = 1.0;
   if (sfd->ToTCPSentOctets.low != 0 &&
         sfd->ToTCPSeqOctets.low != 0)
      ToBLF = (double)sfd->ToTCPSentOctets.low /
         (double)sfd->ToTCPSeqOctets.low;
   if (sfd->FromTCPSentOctets.low != 0 &&
         sfd->FromTCPSeqOctets.low != 0)
      FromBLF = (double)sfd->FromTCPSentOctets.low/
         (double)sfd->FromTCPSeqOctets.low;

   bad_blp = ToBLF < BLP_FRACT || FromBLF < BLP_FRACT;
/*   if ((sfd->flags_seen & 0x0C) == 0x0C) /* Seen both SYNs */
   hdr_printed = 0;
#endif

   lsp = NULL;  sp = sfd->sfc;
   while (sp != NULL) {
      if (gc_call ||  /* Flow was timed out by garbage collector */
            sp->terminating && 
            sp->LastTime < InactTime) {  /* Subflow has timed out */

#if NEW_ATR
	if (!gc_call)  /* Don't count timed-out subflows */
               for (d = fp->distrib_list; d != NULL; d = d->next) {
            switch (d->selector) {
            case FTTCPTIME1:
            case FTTCPTIME2:
               if (sp->ToSeq_seen && sp->FromSeq_seen) {
                  bytes = sp->FromMaxSeq - sp->FromFirstSeq;
                  b2 = sp->ToMaxSeq - sp->ToFirstSeq;
                  if (bytes < b2) bytes = b2;
                  }
               else if (sp->ToSeq_seen)  /* Only 'to' packets seen */
                  bytes = sp->ToMaxSeq - sp->ToFirstSeq;
               else if (sp->FromSeq_seen)  /* Only 'from' packets seen */
                  bytes = sp->FromMaxSeq - sp->FromFirstSeq;
               else bytes = 0;
               break;
            case FTTOTCPSIZE:
            case FTTOTCPRATE1:
            case FTTOTCPRATE2:
               if (!sp->ToSeq_seen) bytes = 0;
               else bytes = sp->ToMaxSeq - sp->ToFirstSeq;
               break;
            case FTFROMTCPSIZE:
            case FTFROMTCPRATE1:
            case FTFROMTCPRATE2:
               if (!sp->FromSeq_seen) bytes = 0;
               else bytes = sp->FromMaxSeq - sp->FromFirstSeq;
               break;
               }
            if (bytes == 0) continue;  /* No data for stream! */
            if (d->LowerSize != 0 || d->UpperSize != 0) {
               n = scale(d, bytes);  /* Divide by 10**Parameter1 */
#define TCPTIME_TEST 0
#if TCPTIME_TEST
               if (d->selector == FTTCPTIME1)
                  printf(" TCPTime1: FromB=%lu|ToB=%lu, n=%u, L=%u|U=%u\n",
                     bytes,b2, n, d->LowerSize,d->UpperSize);
#endif
               if (d->LowerSize != 0 && n <= (Bit32)d->LowerSize) continue;
               if (d->UpperSize != 0 && n > (Bit32)d->UpperSize) continue;
               }

            if (sp->ToSeq_seen && sp->FromSeq_seen) {
               delta = sp->at.LastFromTime >= sp->at.LastToTime ?
                  sp->at.LastFromTime : sp->at.LastToTime;
               if (sp->at.FirstFromTime <= sp->at.FirstToTime)
                  delta -= sp->at.FirstFromTime;
               else delta -= sp->at.FirstToTime;  /* Stream time, us */
               }
            else if (sp->ToSeq_seen)  /* Only 'to' packets seen */
               delta = sp->at.LastToTime >= sp->at.FirstToTime;
            else if (sp->FromSeq_seen)  /* Only 'from' packets seen */
               delta = sp->at.LastToTime >= sp->at.FirstToTime;
            else continue;  /* Didn't see any seq numbers! */

            switch (d->selector) {
            case FTTOTCPSIZE:
            case FTFROMTCPSIZE:
               n = transform(d, bytes);  /* Bytes */
               ++d->counts[n];
               break;
            case FTTCPTIME1:
            case FTTCPTIME2:
               n = transform(d, delta);  /* us */
               ++d->counts[n];
               break;
            case FTTOTCPRATE1:
            case FTTOTCPRATE2:
            case FTFROMTCPRATE1:
            case FTFROMTCPRATE2:
               if (delta == 0) delta = 1;  /* Don't divide by 0 */
               n = transform(d, bytes*1000000.0/delta);  /* B/s */
               ++d->counts[n];
               break;
               }
            }
#endif  /* NEW_ATR */

#if TRACE_NEG_BLP
	 if (bad_blp) {
	    if (!hdr_printed) {
               fprintf(pt_file, "BAD BLP: " 
                  "ToBLF=%.2f [%lu %lu], FromBLF=%.2f [%lu %lu], sfd=%lu <<<<<\n",
                  ToBLF, sfd->ToTCPSentOctets.low, sfd->ToTCPSeqOctets.low,
                  FromBLF, sfd->FromTCPSentOctets.low,sfd->FromTCPSeqOctets.low,
                  sfd);
               hdr_printed = 1;
	       }
            print_pkt_trace(pt_file, "+++", sp,
               "LT=%lu, sp=%lu", sp->LastTime, sp); 
            }
#endif
         terminate_subflow(sfd, sp);
         --sfd->terminating_subflows;  --sfd->active_subflows;

         if (lsp == NULL) sfd->sfc = sp->next_sf;
         else lsp->next_sf = sp->next_sf;
         nsf = sp->next_sf;
         free_subflow(sp);  sp = nsf;

#if TCP_TESTING
         print_subflow_name(stdout, "fin", sp);
         printf(": active=%d, To %lu/%lu, From %lu/%lu",
            sfd->active_subflows,
            sfd->ToTCPSentOctets.low,sfd->ToTCPSeqOctets.low,
            sfd->FromTCPSentOctets.low,sfd->FromTCPSeqOctets.low);
            printf("\n");
#endif
         }
      else {
         lsp = sp;  sp = sp->next_sf;
         }
      }
   }

FILE *pt_file;
#define ptf_title "ntm.trace"

# if TCP_PKT_TRACE
void print_pkt_trace(FILE *f,  char *id,  /* Identifies invocation */
   struct subflow *sfp, char *fmt, ...)  /* Supporting detail */
{
   va_list ap;
   int j,k, seq, mx_seq;
   char *dirn;
   struct sf_trace *sftp = &sfp->sft;
   struct trace_rec *trp;

   print_subflow_name(f, id, sfp);
   va_start(ap, fmt);  vfprintf(f, fmt, ap);  va_end(ap);
   fprintf(f, "\n");

   for (j = sftp->sf_trx, k = 0; k != TRACE_SZ; ++k) {
      trp = &sftp->tr[j];
      if (trp->pkt_nbr != 0 && !(trp->s_flags & sess_PktPrinted)) {
         seq = trp->Seq;  mx_seq = trp->maxSeq;
         if (trp->s_flags & sess_Forward) {
            dirn = "->";
            if (trp->s_flags & ToSYNseen) {
               seq -= sfp->ToFirstSeq;  mx_seq -= sfp->ToFirstSeq;
               }
	    }
         else {  /* Back */
            dirn = "<-";
            if (trp->s_flags & FromSYNseen) {
               seq -= sfp->FromFirstSeq;  mx_seq -= sfp->FromFirstSeq;
               }
	    }
         fprintf(f,
//          "   %5d %7.3f %s %5d %c%c%c%c %lu %lu\n",
            "  %02x %02x  %5d %7.3f %s %5d %c%c%c%c %lu %lu\n",
       trp->flags, trp->s_flags,
            trp->pkt_nbr, trp->time_int/1000.0, dirn, trp->TCPlen,
	    trp->flags & tcp_FlagACK ? 'A' : ' ',
	    trp->flags & tcp_FlagRST ? 'R' : ' ',
	    trp->flags & tcp_FlagSYN ? 'S' : ' ',
	    trp->flags & tcp_FlagFIN ? 'F' : ' ',
            seq, mx_seq);
         trp->s_flags |= sess_PktPrinted;  /* Only print each packet once */
         }
      if (++j == TRACE_SZ) j = 0;
      }
   fflush(f);
   }
# endif

struct subflow *handle_tcp_pkt(struct flow huge *fp,
   int flow_order, struct pkt huge *bp)
{
   struct subflow_name sfn;
   Bit8 *np;
   Bit16 hash;
   int j, tlen;
   struct subflow_data *sfd;
   struct subflow **sf;
   struct subflow *sp, *lsp, *nsf;
   Bit16 ThisFlags;
   Bit32 ThisAck, AckBytes, InactTime;
   Bit32 ThisSeq, SeqBytes;
# if TCP_TRACE
   static Bit32 sf_count = 0;
# endif
# if TCP_PKT_TRACE
   struct trace_rec *sftp;
   double at_ms;
# endif

#define MX_SEQ_DELTA   65535  /* Max TCP window size */
/* #define MX_SEQ_DELTA   29200  /* 20 Ethernet TCP packets */
/* #define MX_SEQ_DELTA   14600  /* 10 Ethernet TCP packets */

   if (flow_order == MREV) {
      memcpy(sfn.SrcPeerAddr,
         bp->High.PeerAddress, sizeof(sfn.SrcPeerAddr));
      memcpy(sfn.SrcTransAddr,
         bp->High.TransAddress, sizeof(sfn.SrcTransAddr));
      memcpy(sfn.DestPeerAddr,
         bp->Low.PeerAddress, sizeof(sfn.DestPeerAddr));
      memcpy(sfn.DestTransAddr,
         bp->Low.TransAddress, sizeof(sfn.DestTransAddr));
      }
   else {
      memcpy(sfn.SrcPeerAddr, 
         bp->Low.PeerAddress, sizeof(sfn.SrcPeerAddr));
      memcpy(sfn.SrcTransAddr,
         bp->Low.TransAddress, sizeof(sfn.SrcTransAddr));
      memcpy(sfn.DestPeerAddr,
         bp->High.PeerAddress, sizeof(sfn.DestPeerAddr));
      memcpy(sfn.DestTransAddr,
         bp->High.TransAddress, sizeof(sfn.DestTransAddr));
      }

   sfn.RuleSet = fp->FlowRuleSet;
   for (np = (Bit8 *)&sfn, hash = 0, j = 0;
          j != sizeof(sfn); ++np, ++j)
      hash += *np * primes[j];
   sf = &sfht[hash %= sfhashmod];
      /* Address of subflow hash table entry */
   if ((sp = *sf) == NULL)
      lsp = NULL;  /* Empty hash chain */
   else {
      for (;;) {
         if (memcmp(&sp->sfn, &sfn, sizeof(sfn)) == 0)
            break;  /* Matched */
         lsp = sp;
         if ((sp = lsp->next_hc) == (struct subflow *)sf) {
            sp = NULL;  break;
            /* Loop stops with lsp -> last subflow in chain */
	    }
         }
      }
   sfd = fp->sfdata;
   if (sp == NULL) {  /* First packet for this hash */
      nsf = get_subflow();
      if (nsf != NULL) {
         if (lsp == NULL) *sf = nsf;  /* Add subflow to hash chain */
         else lsp->next_hc = nsf;
         nsf->next_hc = (struct subflow *)sf;

         nsf->next_sf = sfd->sfc; /* Link subflow into chain for this flow */
         sfd->sfc = nsf;

         memcpy(&nsf->sfn, &sfn, sizeof(sfn));
         ++sfd->n_subflows;  ++sfd->active_subflows;
         if (sfd->active_subflows > sfd->mx_active_subflows)
            sfd->mx_active_subflows = sfd->active_subflows;
         nsf->ToSeq_seen = nsf->FromSeq_seen =
            nsf->flags_seen = nsf->terminating = 0;         
         nsf->ToLastSeq = nsf->FromLastSeq = 0;
         nsf->ToLastLen = nsf->FromLastLen = 0;
# if NEW_ATR
         nsf->at.ToSent = nsf->at.FromSent = 0;
         nsf->at.LastDirection = flow_order;
# endif
# if TCP_TRACE
         nsf->sf_nbr = ++sf_count;
# endif
# if TCP_PKT_TRACE
         nsf->sft.last_pkt_time = microseconds(bp->arrival_time);
         nsf->sft.pkt_nbr = nsf->sft.sf_trx = 0;
         memset(nsf->sft.tr, 0, sizeof(nsf->sft.tr));
# endif
         sp = nsf;

# if TCP_TESTING
         print_subflow_name(stdout, "start", nsf);
         printf("\n");
# endif
         check_subflow_chains(sfd, NULL);
         }
      }

   if (sp != NULL) {  /* We have a subflow */
      ThisFlags = ntohs(bp->th.flags);
      tlen = bp->tcp_len - tcp_dataoffset(ThisFlags);
      ThisAck = ntohl(bp->th.acknum);
      ThisSeq = ntohl(bp->th.seqnum);
# if TCP_PKT_TRACE
      sftp = &sp->sft.tr[sp->sft.sf_trx];
      sftp->pkt_nbr = ++sp->sft.pkt_nbr;
      at_ms = microseconds(bp->arrival_time);
      sftp->time_int = (at_ms - sp->sft.last_pkt_time) / 1000.0;
      sp->sft.last_pkt_time = at_ms;
      sftp->TCPlen = tlen;
      sftp->flags = ThisFlags;
      sftp->s_flags = sp->flags_seen;
      if (flow_order == MREV) {
         sftp->maxSeq = sp->FromMaxSeq;
         }
      else {
         sftp->s_flags |= sess_Forward;
         sftp->maxSeq = sp->ToMaxSeq;
         }
      sftp->Seq = ThisSeq;
      if (++sp->sft.sf_trx == TRACE_SZ) sp->sft.sf_trx = 0;
# endif
# if NEW_ATR
      sp->at.TimeOKrev = 0;
# endif
      if (ThisFlags & tcp_FlagRST)
         sp->flags_seen |= RSTseen;
      else if (flow_order == MREV) {
# if NEW_ATR
         if (sp->at.LastDirection == MFWD) {  /*  'from' after 'to' */
	    if (ThisAck == sp->ToLastSeq + sp->ToLastLen
                  && sp->ToLastLen > 0)
               sp->at.TimeOKrev = 1;
            }
#endif
         if (sp->FromSeq_seen) {
            if (seqcmp(ThisSeq,sp->FromMaxSeq) >= 0) {  /* 'Normal' */
               SeqBytes = ThisSeq - sp->FromMaxSeq;
               if (SeqBytes > 0) {
                  if (SeqBytes <= MX_SEQ_DELTA) {
                     add_Bit32_to_counter64(
                        sfd->FromTCPSeqOctets, SeqBytes);
                     }
                  else {
# if TRACE_SEQ_DECR
                     print_pkt_trace(pt_file, 
                        "FROM-SEQ", sp, ": SeqBytes=%lu <<<", SeqBytes);
# endif
                     }
                  sp->FromMaxSeq = ThisSeq;
                  /* Assume ThisSeq is correct now */
                  }
               }
            else if (seqcmp(ThisSeq,sp->FromLastSeq) < 0) {
               ++sfd->FromTCPDecrSeq;  /* Backward jump in seq */
               }
            }
         else if (ThisSeq != 0) {
# if NEW_ATR
            sp->at.FirstFromTime = microseconds(bp->arrival_time);
# endif
            sp->FromFirstSeq = sp->FromMaxSeq = ThisSeq;
            sp->FromSeq_seen = 1;
            }

         if (tlen > 0) add_Bit32_to_counter64(
            sfd->FromTCPSentOctets, tlen);
         sp->FromLastSeq = ThisSeq;  sp->FromLastLen = tlen;     

#if TCP_PKT_TRACE
         if (ThisFlags & tcp_FlagSYN) sp->flags_seen |= FromSYNseen;
#endif
         if (ThisFlags & tcp_FlagFIN) sp->flags_seen |= FromFINseen;
	 }
      else {
# if NEW_ATR
	 if (sp->at.LastDirection == MREV) {  /* 'to' after 'from' */
	    if (ThisAck == sp->FromLastSeq + sp->FromLastLen
                  && sp->FromLastLen > 0)
               sp->at.TimeOKrev = 1;
            }
# endif
         if (sp->ToSeq_seen) {
            if (seqcmp(ThisSeq,sp->ToMaxSeq) >= 0) {  /* 'Normal' */
               SeqBytes = ThisSeq - sp->ToMaxSeq;
               if (SeqBytes > 0) {
                  if (SeqBytes <= MX_SEQ_DELTA) {
                     add_Bit32_to_counter64(
                        sfd->ToTCPSeqOctets, SeqBytes);
                     }
                  else {
# if TRACE_SEQ_DECR
                     print_pkt_trace(pt_file, 
                        "TO-SEQ", sp, ": SeqBytes=%lu <<<", SeqBytes);
# endif
                     }
                  sp->ToMaxSeq = ThisSeq;
                  /* Assume ThisSeq is correct now */
                  }
               }
            else if (seqcmp(ThisSeq,sp->ToLastSeq) < 0) {
               ++sfd->ToTCPDecrSeq;  /* Backward jump in seq */
               }
            }
         else if (ThisSeq != 0) {
# if NEW_ATR
            sp->at.FirstToTime = microseconds(bp->arrival_time);
# endif
            sp->ToFirstSeq = sp->ToMaxSeq = ThisSeq;
            sp->ToSeq_seen = 1;
            }

         if (tlen > 0) add_Bit32_to_counter64(
            sfd->ToTCPSentOctets, tlen);
         sp->ToLastSeq = ThisSeq;  sp->ToLastLen = tlen;     

#if TCP_PKT_TRACE
         if (ThisFlags & tcp_FlagSYN) sp->flags_seen |= ToSYNseen;
#endif
         if (ThisFlags & tcp_FlagFIN) sp->flags_seen |= ToFINseen;
         }

#if TCP_TRACE
      fprintf(stderr,
      "%5d %9.5f %6.3f %s %s%s len=%4d, ack=%lu, seq=%lu, flg=%04x\n", 
         sp->sf_nbr,
         microseconds(bp->arrival_time)/1000000.0, 
         (microseconds(bp->arrival_time)-
            (flow_order == MREV ? sp->at.LastToTime : sp->at.LastFromTime)
            )/1000000.0,
         flow_order == MREV ? "<-" : "->",
	 flow_order != sp->at.LastDirection ? " " : "=",
         sp->at.TimeOKrev ? "R" : " ",
         tlen, ThisAck, ThisSeq, ThisFlags
         );
#endif

      sp->LastTime = fp->ntm_LastTime;  /* Time last subflow pkt was seen */
      if (!sp->terminating && (sp->flags_seen & RSTseen ||
            (sp->flags_seen & BothFINseen) == BothFINseen)) {
         sp->terminating = 1;
         ++sfd->terminating_subflows;
         }

      if (sfd->terminating_subflows != 0)
         check_tcp_idle_subflows(fp, 0);
      }

   return sp;
   }
#endif  /* TCP_ATR */

struct flow huge *create(struct search_result *s_r,
   int flow_order, struct pkt huge *bp)
{
   struct flow huge *f, huge *ff, huge *lf;
   Bit32 fx;
#if RS_CHAINS
   Bit32 rs_first_flow;
   struct flow huge *pf;
#endif
#if TCP_ATR || defined(MATCH_TRACE)
   int j;
#endif
   if (FloodMode == TV_TRUE)  /* Running in flood mode */
      return NULL;  /* Don't garbage collect for every new flow! */

#if RS_CHAINS
   if ((fx = alloc_flow_nbr()) == 0) {  /* Find a flow data row */
#else
   if ((fx = number_flow()) == 0) {  /* Find a flow data row */
#endif
      bump_noflowpkts(bp->ntm_interface);
      return NULL;  /* No unused rows in flow table */
      }
   f = &fa[fx-1];
#if FLOW_INDEX
   flow_ix[fx-1] = f;  /* 1-org */
#endif
   ++nflows;  /* Nbr of active accounting flows */
   ++FlowsCreated;
   ++ri[c_rtx-1].ri_FlowRecords;  /* 1-org */

   if (f == s_r->lf) {  /* Garbage collector recovered the
               last flow of the hash chain!  (Yes, it can happen) */
      if ((ff = s_r->sf[0]) == NULL) {
         lf = NULL;  /* Empty hash chain */
         }
      else {
         do {
            lf = ff;
            } while ((ff = lf->ht_next) != (struct flow huge *)s_r->sf);
         }  /* Loop stops with lf -> last flow in chain */
      s_r->lf = lf;
      }

   f->ht_next = (struct flow huge *)(s_r->sf);  /* Link f into hash chain */
   if (s_r->lf == NULL) {  /* Empty hash chain */
      s_r->sf[0] = f;  ++n_hash_ents;
      }
   else (s_r->lf)->ht_next = f;

#if RS_CHAINS
   rs_first_flow = ri[c_rtx-1].ri_flow_chain;  /* Link f into rs chain */
   if (rs_first_flow == 0) {  /* First flow for this ruleset */
      f->rs_next = 0;
      ri[c_rtx-1].ri_flow_chain = fx;
      }
   else if (fx < rs_first_flow) {  /* Add to front of chain */
      f->rs_next = rs_first_flow;
      ri[c_rtx-1].ri_flow_chain = fx;
      }
   else {  /* Link into ruleset chain */
      pf = f;  do {
         --pf;
         } while (pf->FlowRuleSet != c_rtx);
      f->rs_next = pf->rs_next;
      pf->rs_next = fx;
      }
# if RS_CH_DEBUG
   check_rs_chain(c_rtx, "create");
# endif
#endif

   memcpy(&f->fk,  /* Copy values from search_key */
      &search_key, sizeof(search_key));
#ifdef MATCH_TRACE
   printf("new flow, hash=%u :  f=%Fp\n", search_hash,f);
   printf("Count: %u+%u",a_stack[0],p_stack[0]);
   for (j = 1; j != p_s_depth; ++j) printf(",%u+%u",a_stack[j],p_stack[j]);
   printf(": ");
   daddr(search_key.Low.PeerAddress);  printf("->");
   daddr(search_key.High.PeerAddress);  printf("\n");
#endif
#if NEW_ATR
   f->distrib_list = NULL;
   f->distrib_bits = 0;
   make_distrib_list(f);
   f->at.ToSent = f->at.FromSent = 0;
   f->at.LastDirection = flow_order == MREV ? MFWD : MREV;
   setzero64(f->LastUpOctets);
   setzero64(f->LastUpPDUs);
   setzero64(f->LastDownOctets);
   setzero64(f->LastDownPDUs);
#endif
   setzero64(f->UpOctets);
   setzero64(f->UpPDUs);
   setzero64(f->DownOctets);
   setzero64(f->DownPDUs);
   f->FlowRuleSet = c_rtx;
#if NF_CISCO_DATA
   f->LastTime = f->FirstTime = bp->dFirst;
#else
   f->FirstTime = centiseconds(bp->arrival_time);
#endif
#if TCP_ATR
   make_tcp_data(f);
#endif
   return f;
   }

#if NEW_ATR
static double pwr_of_10[10] = {
   1.0,
   0.1,
   0.01,
   0.001,
   0.0001,
   0.00001,
   0.000001,
   0.0000001,
   0.00000001,
   0.000000001 };

#if TCP_ATR
Bit32 scale(struct distribution *d, double y)
{
   double x;
   x = y*pwr_of_10[d->ScaleFactor > 9 ? 9 : d->Parameter1];
   return (Bit32)floor(x);
   }
#endif  /* TCP_ATR */

int transform(struct distribution *d, double y)
{
   double x, T;
   x = y*pwr_of_10[d->ScaleFactor > 9 ? 9 : d->ScaleFactor];
   if (x <= (double)d->LowerLimit) return 0;
   else if (x > (double)d->UpperLimit) {
      return d->Buckets;
      }
   else switch (d->Transform) {
   case DS_LIN:
      T = ((double)(x - (double)d->LowerLimit)) * d->M;
      break;
   case DS_LOG:
      T = log10((double)x/(double)d->LowerLimit) * d->M;
      break;
      }
   return 1 + (int)floor(T);
   }
#endif

void count(struct flow huge *fp,
   int flow_order, struct pkt huge *bp)
{
   struct search_result *s_r;
#if TCP_ATR 
   struct subflow *sfp;
#endif
#if NEW_ATR
   struct distribution *d;
   struct at_data *atp;
   double at_us, delta_us;
   int n;
#endif
#if NF_CISCO_DATA
   if (flow_order == MREV) {
      add_Bit32_to_counter64(fp->DownOctets,bp->dOctets);
      add_Bit32_to_counter64(fp->DownPDUs,bp->dPkts);
      }
   else {
      add_Bit32_to_counter64(fp->UpOctets,bp->dOctets);
      add_Bit32_to_counter64(fp->UpPDUs,bp->dPkts);
      }
   if (bp->dFirst < fp->FirstTime)
      fp->FirstTime = bp->dFirst;
   if (bp->dLast > fp->LastTime)
      fp->LastTime = bp->dLast;
   fp->ntm_LastTime = uptime_cs();  /* NetFlowMet meter time! */
#else
   if (flow_order == MREV) {
      add_Bit32_to_counter64(fp->DownOctets,bp->p_len);
      incr_counter64(fp->DownPDUs);
      }
   else {
      add_Bit32_to_counter64(fp->UpOctets,bp->p_len);
      incr_counter64(fp->UpPDUs); 
     }
   fp->ntm_LastTime = centiseconds(bp->arrival_time);
      /* Also sets LastTime (see flowhash.h) */
#endif

#if TCP_ATR 
   sfp = fp->sfdata ? handle_tcp_pkt(fp, flow_order, bp) : NULL;
#endif  /* TCP_ATR */

#if NEW_ATR && !NF_CISCO_DATA
   at_us = microseconds(bp->arrival_time);
# if TCP_ATR
   atp = sfp ? &sfp->at : &fp->at;
# else
   atp = &fp->at;
# endif
   for (d = fp->distrib_list; d != NULL; d = d->next) {
      if (flow_order == MREV) {
         switch (d->selector) {
         case FTFROMPACKETSIZE:
            n = transform(d, bp->p_len);
            ++d->counts[n];
            break;
         case FTFROMINTERARRIVALTIME:
            if (atp->FromSent) {
               delta_us = at_us - atp->LastFromTime;
               n = transform(d, delta_us);
               ++d->counts[n];
               }
            break;
         case FTTOTURNAROUNDTIME:
            if (
#if TCP_ATR
                  sfp && atp->TimeOKrev || !sfp &&
#endif
                  atp->ToSent && flow_order != atp->LastDirection) {
               delta_us = at_us - atp->LastToTime;
               n = transform(d, delta_us);
               ++d->counts[n];
               }
            break;
            }
         }
      else {
         switch (d->selector) {
         case FTTOPACKETSIZE:
            n = transform(d, bp->p_len);
            ++d->counts[n];
            break;
         case FTTOINTERARRIVALTIME:
            if (atp->ToSent) {
               delta_us = at_us - atp->LastToTime;
               n = transform(d, delta_us);
               ++d->counts[n];
               }
            break;
         case FTFROMTURNAROUNDTIME:
            if (
#if TCP_ATR
                  sfp && atp->TimeOKrev || !sfp &&
#endif
                  atp->FromSent && flow_order != atp->LastDirection) {
               delta_us = at_us - atp->LastFromTime;
               n = transform(d, delta_us);
               ++d->counts[n];
               }
            break;
            }
         }
      }

   if (flow_order == MREV) {  /* Remember time for last packet */
      atp->LastFromTime = at_us;
      atp->FromSent = 1;
      }
   else {
      atp->LastToTime = at_us;
      atp->ToSent = 1;
      }
   atp->LastDirection = flow_order;
#endif
   }

void pkt_monitor(struct pkt huge *bp)
{
   int x, r;
   struct rtinf_rec *rip;
   struct flow huge *f;

#if defined(DOS)
   if (bp->PeerAddrType == AT_DUMMY) {
      if (++dummypackets == 1000) {
         ++kilodummypackets;  dummypackets = 0;
         }
      ++dummypacketrate;
      }
#endif

   for (x = 0; x != sizeof(ri)/sizeof(struct rtinf_rec); ++x) {
      rip = &ri[x];
      if (rip->ri_running == 0) continue;

      c_rtx = x+1;  /* Rule table index of current rule table */
      c_rt = rip->ri_rule_table;  /* Current Rule Table */
      c_rht = rip->ri_rule_hash;  /* Current Rule Hash Table */
      c_rsz = rip->ri_Size;  /* Nbr of rules in current rule table */

      if ((r = match(1, bp, &bp->Low,&bp->High)) == MIGNORE)
         continue;  /* Try next rule set */
         /* match() builds search_key from bp */
      else if (r == MFAIL) {  /* S->D Fail */
         if ((r = match(0, bp, &bp->High,&bp->Low)) != MSUC)
            continue;  /* Ignore or fail */
         else if ((f = current(&ds_pointers,
               &search_key.Low,&search_key.High)) != NULL)
            count(f,MREV,bp);
         else if ((f = current(&sd_pointers,
               &search_key.High,&search_key.Low)) != NULL)
            count(f,MFWD,bp);
         else {
            f = create(&ds_pointers,MREV,bp);
            if (f == NULL) continue;  /* Coudn't create flow */
            count(f,MREV,bp);
            }
         }
      else {  /* S->D Suc */
         if ((f = current(&sd_pointers,
               &search_key.Low,&search_key.High)) != NULL)
            count(f,MFWD,bp);
         else if ((f = current(&ds_pointers,
               &search_key.High,&search_key.Low)) != NULL)
            count(f,MREV,bp);
         else {
            f = create(&sd_pointers,MFWD,bp);
            if (f == NULL) continue;  /* Coudn't create flow */
            count(f,MFWD,bp);
            }
         }
      }
   }

#define MAXPKTLEN  1526

#if !NF_CISCO_DATA  /* Front-end calls these routines to get attrib values */

#ifdef DECNET
int dn_node(Bit8 huge *dn_addr)
{
   return dn_addr[1]<<8 | dn_addr[0];
   }

#define dn_area(node) (node >> 10)
#define dn_host(node) (node & 0x3FF)
#endif

void unpack_dn_node(Bit8 huge *ap, Bit8 huge *dn_addr)
{
   Bit8 dl = dn_addr[1];
   ap[0] = dl >> 2;    /* DECnet area */
   ap[1] = dl & 0x03;  ap[2] = dn_addr[0];  /* DECnet host */
   ap[3] = 0;
   }

void unpack_at_node(Bit8 huge *ap, Bit8 huge *at_net, Bit8 at_node)
{
   ap[0] = at_net[0];  ap[1] = at_net[1];
   ap[2] = at_node;
   ap[3] = 0;
   }

#define BAD_ADJ    1
#define BAD_PEER   2
#define BAD_TRANS  3

void zero_pkt_fields(int which, struct pkt huge *bp)
{
   switch (which) {
   case BAD_ADJ:
      memset(bp->Low.AdjAddress, 0, 
         MAC_ADDR_LEN+PEER_ADDR_LEN+TRANS_ADDR_LEN);
      memset(bp->High.AdjAddress, 0, 
         MAC_ADDR_LEN+PEER_ADDR_LEN+TRANS_ADDR_LEN);
      break;
   case BAD_PEER:
      memset(bp->Low.PeerAddress,  0, PEER_ADDR_LEN+TRANS_ADDR_LEN);
      memset(bp->High.PeerAddress, 0, PEER_ADDR_LEN+TRANS_ADDR_LEN);
      break;
   case BAD_TRANS:
      memset(bp->Low.TransAddress,  0, TRANS_ADDR_LEN);
      memset(bp->High.TransAddress, 0, TRANS_ADDR_LEN);
      break;
      }
   }

#if V6
int find_v6_type(int *prot, int *hx, Bit8 *p, int hl)
{
   int nh, nx;

   nh = p[6];  nx = 40;
   for (;;) {
      switch (nh) {
      case PT_6FRAG:  /* Fragment */
      case PT_6ESP:   /* Encapsulating Security Payload (ESP) RFC 2406 */
      case PT_6NONH:  /* No next header */
         *prot = nh;
         return 0;  /* Can't see into this packet */

      /* case PT_6ICMP:  /* ICMPv6 */
      default:
         *prot = nh;  *hx = nx;
         return 1;  /* Can see into this one */

      case PT_6HOP:   /* Hop-by-hop options */
      case PT_6ROUT:  /* Routing (Type 0) */
      case PT_6DEST:  /* Destination options */
      case PT_6AH:    /* Authentication (AH) RFC 2402 */
         break;  /* Keep looking */
         }
      if (hl < nx) {
         *prot = -1;  /* Not enough header bytes to get Next Header */
         return 0;
         }
      nh = p[nx];
      nx += (p[nx]+1) * 8;
      }
   }
#endif

void pkt_extract(struct pkt huge *bp,
   int type, Bit8 huge *hdrp, int hl)
{
   int prot, dtype, dx, j;
   union decnet huge *dp;
#ifdef DECNET
   int dlen, snode, dnode, area, host;
#endif
#if NF_OCX_BGP
   Subnet *my_subnet;
   Bit32 my_IP4addr;
   Bit16 my_ASN;
#endif

   bp->PeerAddrType = type;
   if (hl < 0) {  /* Didn't get the header! */
      zero_pkt_fields(BAD_ADJ, bp);
      return;
      }

   switch (type) {
#if V6
   case AT_IP6:
      bp->DSCodePoint =  /* Traffic Class  field, RFC 2460 */
         (hdrp[1] >> 6) | ((hdrp[0] & 0x0F) << 2);
      if (hl < 8+IP6_ADDR_LEN*2) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, bp);
         return;
         }
      memcpy(bp->Low.PeerAddress,&hdrp[8],IP6_ADDR_LEN);  /* IPv6 source */
      memcpy(bp->High.PeerAddress,&hdrp[24],IP6_ADDR_LEN);  /* IPv6 dest */
      if (!find_v6_type(&prot,&dx, hdrp,hl)) {
         if (prot < 0)  /* Not enough bytes to determine protocol */
            bp->TransAddrType = 0;
	else  /* Can't see into packet body */
            bp->TransAddrType = prot;
         zero_pkt_fields(BAD_TRANS, bp);
         return;
         }

      switch (bp->TransAddrType = prot) {
      case PT_UDP:
      case PT_TCP:
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, bp);
            return;
            }
         bp->Low.TransAddress[0] = hdrp[dx];  /* Source port */
         bp->Low.TransAddress[1] = hdrp[dx+1];
         bp->High.TransAddress[0] = hdrp[dx+2];  /* Dest port */
         bp->High.TransAddress[1] = hdrp[dx+3];
         break;
      case PT_6ICMP:  /* RFC 2463 */
      case PT_ICMP:
         if (hl < dx+2) {  /* Didn't get the type,code fields */
            zero_pkt_fields(BAD_TRANS, bp);
            return;
            }
         bp->Low.TransAddress[0] = bp->High.TransAddress[0] = 0;
         bp->Low.TransAddress[1] = hdrp[dx];  /* Type field */
         bp->High.TransAddress[1] = hdrp[dx+1];  /* Code field */
         break;
         }
#if TCP_ATR
      if (prot == PT_TCP) {
         if (hl < dx+4+sizeof(bp->th)) {  /* Didn't get the tcp header */
            bp->tcp_len = 0;
            memset(&bp->th, 0, sizeof(bp->th));
            }
         else {
            bp->tcp_len = (hdrp[4] << 8 | hdrp[5]) - dx;
            memcpy(&bp->th, &hdrp[dx+4], sizeof(bp->th));
	    }
         }
#endif  /* TCP_ATR */
      return;
#endif
   case AT_IP4:
      bp->DSCodePoint = hdrp[1] >> 2;  /* Old TOS field, RFC 2474 */
      if (hl < 12+IP4_ADDR_LEN*2) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, bp);
         return;
         }
      memcpy(bp->Low.PeerAddress,&hdrp[12],IP4_ADDR_LEN);  /* IPv4 source */
      memcpy(bp->High.PeerAddress,&hdrp[16],IP4_ADDR_LEN);  /* IPv4 dest */
#if V6
      memset(&bp->Low.PeerAddress[IP4_ADDR_LEN], 0, 
         PEER_ADDR_LEN-IP4_ADDR_LEN);
      memset(&bp->High.PeerAddress[IP4_ADDR_LEN], 0, 
         PEER_ADDR_LEN-IP4_ADDR_LEN);
#endif
      prot = bp->TransAddrType = hdrp[9];  /* IPv4 Protocol Type */
      if (hdrp[6] & 0x1F != 0 || hdrp[7] != 0) {  /* Fragment offset != 0 */
         zero_pkt_fields(BAD_TRANS, bp);  /* Can't see into fragment */
         return;
         }
      dx = (hdrp[0] & 0x0F) << 2;  /* HLEN */
      /* Data offset.  HLEN = IPv4 header length in 32-bit units */
      if (prot == PT_ICMP) {
         if (hl < dx+2) {  /* Didn't get the type,code fields */
            zero_pkt_fields(BAD_TRANS, bp);
            return;
            }
         bp->Low.TransAddress[0] = bp->High.TransAddress[0] = 0;
         bp->Low.TransAddress[1] = hdrp[dx];  /* Type field */
         bp->High.TransAddress[1] = hdrp[dx+1];  /* Code field */
         }
      else {
         if (hl < dx+4) {  /* Didn't get the IP ports! */
            zero_pkt_fields(BAD_TRANS, bp);
            return;
            }
         bp->Low.TransAddress[0] = hdrp[dx];  /* Source port */
         bp->Low.TransAddress[1] = hdrp[dx+1];
         bp->High.TransAddress[0] = hdrp[dx+2];  /* Dest port */
         bp->High.TransAddress[1] = hdrp[dx+3];
         }
#if TCP_ATR
      if (prot == PT_TCP) {
         if (hl < dx+4+sizeof(bp->th)) {  /* Didn't get the tcp header */
            bp->tcp_len = 0;
            memset(&bp->th, 0, sizeof(bp->th));
            }
         else {
            bp->tcp_len = (hdrp[2] << 8 | hdrp[3]) - dx;
            memcpy(&bp->th, &hdrp[dx+4], sizeof(bp->th));
	    }
         }
#endif  /* TCP_ATR */
#if NF_OCX_BGP
#define get32(a)  ((a[0] << 8 | a[1]) << 8 | a[2]) << 8 | a[3]
      if (asn_lookup) {
         my_IP4addr = get32(bp->Low.PeerAddress);
         if ((my_subnet = FindSubnet(ntohl(my_IP4addr))) != NULL) {
            my_ASN = use_owner_asns ? my_subnet->src_as : my_subnet->exit_as;
            bp->Low.nf_ASN[0] = my_ASN >> 8;
            bp->Low.nf_ASN[1] = my_ASN & 0xff;
            bp->Low.nf_mask = bits_in_mask(my_subnet->mask);
            }
         my_IP4addr = get32(bp->High.PeerAddress);
         if ((my_subnet = FindSubnet(ntohl(my_IP4addr))) != NULL) {
            my_ASN = use_owner_asns ? my_subnet->src_as : my_subnet->exit_as;
            bp->High.nf_ASN[0] = my_ASN >> 8;
            bp->High.nf_ASN[1] = my_ASN & 0xff;
            bp->High.nf_mask = bits_in_mask(my_subnet->mask);
            }
         }
#endif
      return;
   case AT_NOVELL:
      bp->TransAddrType = hdrp[5];  /* XNS packet type */
      if (hl < 18+IPX_ADDR_LEN) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, bp);
         return;
         }
      memcpy(bp->High.PeerAddress,&hdrp[6],IPX_ADDR_LEN);  /* IPX dest net */
      memcpy(bp->Low.PeerAddress,&hdrp[18],IPX_ADDR_LEN);  /* IPX soure net */
      if (hl < 30) {  /* Didn't get the IPX sockets! */
         zero_pkt_fields(BAD_TRANS, bp);
         return;
         }
      bp->High.TransAddress[0] = hdrp[16];
      bp->High.TransAddress[1] = hdrp[17];  /* IPX dest socket */
      bp->Low.TransAddress[0] = hdrp[28];
      bp->Low.TransAddress[1] = hdrp[29];  /* IPX source socket */
      return;
   case AT_ETHERTALK:
      if (hl < 10) {  /* Didn't get the peer addresses! */
         zero_pkt_fields(BAD_PEER, bp);
         return;
         }
      unpack_at_node(bp->Low.PeerAddress, &hdrp[6],hdrp[9]);  /* AT source */
      unpack_at_node(bp->High.PeerAddress, &hdrp[4],hdrp[8]);  /* AT source */
      if (hl < 13) {  /* Didn't get the AT sockets! */
         zero_pkt_fields(BAD_TRANS, bp);
         return;
         }
      bp->TransAddrType =hdrp[12];  /* AT DDP protocol type */
      bp->Low.TransAddress[0] = bp->High.TransAddress[0] = 0;
      bp->Low.TransAddress[1] = hdrp[11];  /* AT source socket */
      bp->High.TransAddress[1] = hdrp[10];  /* AT dest socket */
      return;
   case AT_CLNS:
      bp->TransAddrType = hdrp[5] & 0x1F;  /* CLNS packet type */
      bp->Low.TransAddress[0] = bp->High.TransAddress[0] =
         bp->Low.TransAddress[1] = bp->High.TransAddress[1] = 0;
      memcpy(bp->High.PeerAddress,&hdrp[11],j = hdrp[10]);  /* Dest */
      if (j < NSAP_ADDR_LEN)
         memset(bp->High.PeerAddress + j, 0, NSAP_ADDR_LEN-j);
      memcpy(bp->Low.PeerAddress,&hdrp[12+j],dx = hdrp[11+j]);  /* Source */
      if (dx < NSAP_ADDR_LEN)
         memset(bp->Low.PeerAddress + dx, 0, NSAP_ADDR_LEN-dx);
      return;
   case AT_DECNET:
      bp->Low.TransAddress[0] = bp->High.TransAddress[0] =
         bp->Low.TransAddress[1] = bp->High.TransAddress[1] = 0;
      dtype = hdrp[2];  /* DECnet packet type */
      if (dtype != 0x81) dp = (union decnet huge *)&hdrp[3];
      else {
         dtype = hdrp[3];
         dp = (union decnet huge *)&hdrp[4];
         }
      switch (bp->TransAddrType = dtype & 0x0F) {
      case 0x06:  /* Data */
      case 0x0E:  /* Data + discard flag */
         unpack_dn_node(bp->Low.PeerAddress, dp->d.src_dn_addr);
         unpack_dn_node(bp->High.PeerAddress, dp->d.dest_dn_addr);
#ifdef DECNET
         dnode = dn_node(dp->d.dest_dn_addr);
         snode = dn_node(dp->d.src_dn_addr);
         fprintf(logfl,"Data to %d.%d from %d.%d\n",
            dn_area(dnode),dn_host(dnode),
            dn_area(snode),dn_host(snode) );
#endif
         break;
      case 0x07:  /* Level 1 routing */
         unpack_dn_node(bp->Low.PeerAddress, dp->l1r.src_dn_addr);
#ifdef DECNET
         snode = dn_node(dp->l1r.src_dn_addr);
         fprintf(logfl,"Level 1 routing from %d.%d\n",
            dn_area(snode),dn_host(snode));
#endif
         break;
      case 0x09:  /* Level 2 routing */
         unpack_dn_node(bp->Low.PeerAddress, dp->l2r.src_dn_addr);
#ifdef DECNET
         snode = dn_node(dp->l2r.src_dn_addr);
         fprintf(logfl,"Level 2 routing from %d.%d\n",
            dn_area(snode),dn_host(snode));
#endif
         break;
      case 0x0B:  /* Router hello */
         unpack_dn_node(bp->Low.PeerAddress, dp->rh.src_dn_addr);
#ifdef DECNET
         snode = dn_node(dp->rh.src_dn_addr);
         dnode = dn_node(dp->rh.rtr_dn_addr);
         fprintf(logfl,"Router hello from %d.%d, other router %d.%d\n",
            dn_area(snode),dn_host(snode),
            dn_area(dnode),dn_host(dnode) );
#endif
         break;
      case 0x0D:  /* Endnode hello */
         unpack_dn_node(bp->Low.PeerAddress, dp->eh.src_dn_addr);
#ifdef DECNET
         snode = dn_node(dp->eh.src_dn_addr);
         dnode = dn_node(dp->eh.rtr_dn_addr);
         fprintf(logfl,"Endnode hello from %d.%d, designated router %d.%d\n",
            dn_area(snode),dn_host(snode),
            dn_area(dnode),dn_host(dnode) );
#endif
         break;
      default:  /* Unknown DECnet type */
         scpos(0,scr_lrow);
         printf("\nDN pkt type %02x: ", dtype);
         for (j=0; j != 16; ++j) printf(" %02x",hdrp[j]);
         printf("\n");
         for (j=16; j != 34; ++j) printf(" %02x",hdrp[j]);
         printf("\n");
#ifdef TESTING
         if (!logfl) open_log();
         fprintf(logfl, "\nDN pkt type %02x: ", dtype);
         for (j=17; j != 34; ++j) fprintf(logfl," %02x",hdrp[j]);
         fprintf(logfl, "\n");
#endif
         }
      return;
      }
   }

#define putshort(a,n)  \
   a[0] = (n >> 8) & 0xFF;  a[1] = n & 0xFF;

void other_extract(struct pkt huge *bp,
   Bit16 ethertype, Bit16 lsap)
{
   bp->PeerAddrType = AT_OTHER;
   putshort(bp->Low.PeerAddress,ethertype);  /* Blue Book type */
   putshort(bp->High.PeerAddress,lsap);  /* 802.2 LSAP */
   bp->TransAddrType =
      bp->Low.TransAddress[0] = bp->Low.TransAddress[1] =
      bp->High.TransAddress[0] = bp->High.TransAddress[1] = 0;
   }

int handle_pkt(struct pkt huge *pp,
   Bit16 ether_type, Bit16 lsap,
   Bit8 huge *ethp, 
   Bit8 huge *p, int pl)
{
   pp->PeerAddrType = AT_DUMMY;
#if V6 && V6_TESTV4
   memset(pp->Low.PeerAddress,0,PEER_ADDR_LEN);
   memset(pp->High.PeerAddress,0,PEER_ADDR_LEN);
#endif
   switch (ether_type) {
   case 0x0800:
#if V6 && V6_TESTV4
      if (proto_reqd[AT_IP6]) {
         pkt_extract(pp, AT_IP4, p,pl);
         if (use_ip_length) pp->p_len = p[2] << 8 | p[3];
         memcpy(&pp->Low.PeerAddress[12],&pp->Low.PeerAddress[0],4);
         memcpy(&pp->High.PeerAddress[12],&pp->High.PeerAddress[0],4);
         memset(pp->Low.PeerAddress,0,4);
         memset(pp->High.PeerAddress,0,4);
         pp->PeerAddrType = AT_IP6;
         }
      else if (proto_reqd[AT_IP4]) {
         pkt_extract(pp, AT_IP4, p,pl);
         if (use_ip_length) pp->p_len = p[2] << 8 | p[3];
         }
#elif V6
      if ((p[0] & 0xF0) == 0x40 && proto_reqd[AT_IP4]) {
         pkt_extract(pp, AT_IP4, p,pl);
         if (use_ip_length) pp->p_len = p[2] << 8 | p[3];
         }
      else if ((p[0] & 0xF0) == 0x60 && proto_reqd[AT_IP6]) {
         pkt_extract(pp, AT_IP6, p,pl);
         if (use_ip_length) pp->p_len = p[4] << 8 | p[5];
         }
#else
      if (proto_reqd[AT_IP4]) {
         pkt_extract(pp, AT_IP4, p,pl);
         if (use_ip_length) pp->p_len = p[2] << 8 | p[3];
         }
#endif
      break;
   case 0x809B:
      if (proto_reqd[AT_ETHERTALK]) pkt_extract(pp, AT_ETHERTALK, p,pl);
      break;
   case 0x8137:
      if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p,pl);
      break;
   case 0x6003:
      if (proto_reqd[AT_DECNET]) pkt_extract(pp, AT_DECNET, p,pl);
      break;
   case 0x0000:  /* 802.3 packet other than SNAP */
      switch (lsap) {
      case 0xFFFF:
         if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p,pl);
         break;  /* Novell "Raw 802.2" doesn't use LLC */
#ifdef CLNS
      case 0xFEFE:  /* CLNS */
         if (proto_reqd[AT_CLNS]) pkt_extract(pp, AT_CLNS, p+2,pl-2);
         break;  /* ISO 8473-1 uses 1-org, byte 0 is LLC control byte */
#endif
      case 0xE0E0:  /* Novell 802.2 */
         if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p+3,pl-3);
         break;
/*    case 0xF0F0:  NetBIOS */
         }
      }
   if (proto_reqd[AT_OTHER] && pp->PeerAddrType == AT_DUMMY)
      other_extract(pp, ether_type,lsap);

   if (pp->PeerAddrType != AT_DUMMY && adj_reqd) {
      memcpy(pp->Low.AdjAddress,ethp+6,MAC_ADDR_LEN);  /* Source */
      memcpy(pp->High.AdjAddress,ethp,MAC_ADDR_LEN);  /* Dest */
      }

   return pp->PeerAddrType != AT_DUMMY;
   }

#endif  /* NF_ASN_ATT */

#ifdef DOS
void save_time(void)
{
   s_tod_h = tod_h;  s_tod_m = tod_m;  s_tod_s = tod_s;
   elapsed_sec = 0;
   }
#endif

void show_memory(void)
{
   char msg[60];

   display_msg(1,"Memory Usage Info ..");
   sprintf(msg, "Rules: %lu max, %lu free", 
      mxrules, mxrules-inuse_rules);
   display_msg(0,msg);
   sprintf(msg, "RuleHash: %lu max, %lu free", 
      mxrulehash, mxrulehash-inuse_rule_hash);
   display_msg(0,msg);
   sprintf(msg, "Flows: %lu max (%d B/flow)", 
      mxflows, sizeof(struct flow));
   display_msg(0,msg);
   sprintf(msg, "FlowHash: %lu slots",
      fthashmod);
   display_msg(0,msg);
#if TCP_ATR
   sprintf(msg, "TCP streams: %lu max, %lu free",
      mxsubflows, n_free_subflows);
   display_msg(0,msg);
   sprintf(msg, "TCP flows: %lu alloc, %lu free (%d B/flow)",
      mxsfd_blocks, n_free_sfd_blocks, sizeof(struct subflow_data));
   display_msg(0,msg);
#endif
#if NEW_ATR
   sprintf(msg, "Distribs: %lu max, %lu free (%d B/dist)",
      mxdistribs, n_free_distribs, sizeof(struct distribution));
   display_msg(0,msg);
   sprintf(msg, "Distrib Events: %lu alloc, %lu free",
      mxdistevents, n_free_events);
   display_msg(0,msg);
#endif

#if defined(DOS)
   sprintf(msg, "malloc(): %lu bytes free", coreleft());
   display_msg(0,msg);
#endif
#if FLAT_MEM
   sprintf(msg, "flat: %lu bytes free", flat_remaining());
   display_msg(0,msg);
#endif
   }

#if RS_CHAINS
void show_chains(int reset)
{
   char msg[60];

   display_msg(1,"Ruleset Chain Info ..");
   sprintf(msg, "%d tests", rsc_n_searches);
   display_msg(0,msg);
   if (rsc_n_searches == 0) return;
   sprintf(msg, "Range: %lu max, %lu av", 
      rsc_mx_range, rsc_av_range/rsc_n_searches);
   display_msg(0,msg);
   sprintf(msg, "Tests: %lu max, %lu av", 
      rsc_mx_tests, rsc_av_tests/rsc_n_searches);
   display_msg(0,msg);

   if (reset) {
      rsc_n_searches = 
         rsc_mx_range = rsc_av_range =
         rsc_mx_tests = rsc_av_tests = 0;
      }
   }
#endif

Bit32 active_flows(void)
{
   Bit32 af,j;
   struct flow huge *fp;
   Bit32 GCtime;
   for (af = 0, j = 1; j <= mxflows; ++j) {
      /* Flow indexes from 1 to MXFLOWS */
      if (flow_idle(j)) continue;
#if FLOW_INDEX
      fp = flow_ix[j-1];
#else
      fp = &fa[j-1];
#endif
      GCtime = ri[fp->FlowRuleSet-1].ri_gc_time;
      if (fp->ntm_LastTime > GCtime) ++af;  /* Can't be recovered */
      }
   return af;
   }

void show_flow_table(void)
{
   char msg[60];
   Bit32 f;
   display_msg(1,"Flow Table Info ..");
   sprintf(msg, "InactTime %d, Flood %d",
      InactivityTimeout,FloodMark);
   display_msg(0,msg);
   sprintf(msg, "Flows: %lu active, %u used",
      active_flows(), nflows);
   display_msg(0,msg);
   sprintf(msg,"  %lu recovered (GC: %u s, %u flow)",
      FlowsRecovered, gc_interval,gc_f);
   display_msg(0,msg);
   if (stats_time != 0) {
      f = FlowsCreated*1000L/stats_time;
      sprintf(msg,"  Creating   %lu.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      f = FlowsRecovered*1000L/stats_time;
      sprintf(msg,"  Recovering %lu.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      }
   }

void show_protocols(void)
{
   int j;
   char msg[60];
   display_msg(1,"Protocols being metered ..");
   sprintf(msg,"      %3.3s  %s",  adj_reqd ? "Yes" : " No", "Adjacent");
   display_msg(0,msg);
#if NF_OCX_BGP
   sprintf(msg,"      %3.3s  %s",  asn_lookup ? "Yes" : " No", "ASN lookup");
   display_msg(0,msg);
#endif
   for (j = 0; j != sizeof(proto_reqd)/sizeof(int); ++j) {
      if (proto_names[j][0] != '\0') {
         sprintf(msg,"  %2d  %3.3s  %s",
            j, proto_reqd[j] ? "Yes" : " No", proto_names[j]);
         display_msg(0,msg);
         }
      }
   }

void zero_stats(void)
{
   FlowsCreated = FlowsRecovered =
      n_matches = n_hash_compares = n_hash_searches = 0L;
   clear_pkt_stats = 1;
#ifdef DOS
   scpos(0,scr_lrow);  printf("Statistics Zeroed");
#else
   display_msg(0,"Statistics Zeroed");
#endif
   }

void show_stats(void)
{
   int i,j;
   Bit32 aps,apb, kdp;
   char msg[60];
#ifdef DOS
   if (stats_time == 0 || npackets == 0 ||
     kilodummypackets == 0 || mindummyrate == 0)
#else
   if (stats_time == 0 || npackets == 0)
#endif
      return;  /* Avoid divides by zero */

   display_msg(1,"Meter Statistics ..");
   aps = (npackets*10+5L)/(stats_time*10L);
#ifdef DOS
   apb = (t_backlog*10+5L)/(stats_time*10L);
   sprintf(msg,"Av pkt/s %lu, av pkt backlog %lu", aps,apb);
   display_msg(0,msg);
   sprintf(msg,"Max pkt/s %u, max pkt backlog %u",
      max_pkt_rate,max_pkt_backlog);
   display_msg(0,msg);
   if (kilodummypackets != 0) {
      if (dummypackets >= 500) ++kilodummypackets;
      i = kilodummypackets*1000L/(kilodummypackets+npackets/1000L);
      j = (mindummyrate*10000L+5L)/((mindummyrate+mdpacketrate)*10L);
      sprintf(msg,"Idle time av %u.%u, min %u.%u %", i/10,i%10, j/10,j%10);
      display_msg(0,msg);
      }
#else
   sprintf(msg,"Av pkt/s %lu, max pkt/s %u", aps,max_pkt_rate);
   display_msg(0,msg);
#endif
   sprintf(msg,"%lu flows active (max %lu)",active_flows(),mxflows);
   display_msg(0,msg);
   sprintf(msg,"%u masks in table (max %u)", n_masks,MXMASKS);
   display_msg(0,msg);
   i = (n_matches*100L+5L)/(npackets*10L);
   j = (n_hash_searches*100L+5L)/(npackets*10L);
   sprintf(msg,"%u.%u rules/pkt, %u.%u searches/pkt",
      i/10,i%10, j/10,j%10);
   display_msg(0,msg);
   if (n_hash_searches != 0) {
      i = (n_hash_compares*100L+5L)/(n_hash_searches*10L);
      sprintf(msg,"%u.%u compares/search", i/10,i%10);
      display_msg(0,msg);
      sprintf(msg,"%u hash slots, %u in use, ", t_hash_size,n_hash_ents);
      }
   }

void show_rulesets(void)
{
   int c,j,t;
   struct rtinf_rec *rip;
   char buf[80];
   display_msg(1,"Ruleset Table ..");
   for (c = t = j = 0; j != sizeof(ri)/sizeof(struct rtinf_rec); ++j) {
      rip = &ri[j];
      if (rip->ri_Status != RS_ACTIVE &&
            rip->ri_Status != RS_NOTINSERVICE
         ) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"   Rdy  Flows Rules Name  Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %c  %5lu  %5u  %s  %s",
         j+1,  /* 1-org */
         rip->ri_Status == RS_ACTIVE ? 'T' : 'F',
         rip->ri_FlowRecords,
         rip->ri_Size, rip->ri_Name,
         rip->ri_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No ruleset rows active");
   }

void show_managers(void)
{
   int c,j,t;
   struct mgr_rec *mip;
   char buf[80];
   display_msg(1,"Manager Table ..");
   for (c = t = j = 0; j != sizeof(mi)/sizeof(struct mgr_rec); ++j) {
      mip = &mi[j];
      if (mip->mi_Status != RS_ACTIVE) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"  Crnt Stby HWM RSby  Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %2d  %2d  %3d   %c  %s",
         j+1,  /* 1-org */
         mip->mi_CurrentRuleSet, mip->mi_StandbyRuleSet,
         mip->mi_HighWaterMark,
         mip->mi_RunningStandby == TV_TRUE ? 'T' : 'F',
         mip->mi_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No manager rows active");
   }

void show_readers(void)
{
   int c,j,t;
   struct rdr_rec *cip;
   Bit32 now;
   char buf[80];
   display_msg(1,"Meter Reader Table ..");
   now = uptime_cs();
   for (c = t = j = 0; j != sizeof(ci)/sizeof(struct rdr_rec); ++j) {
      cip = &ci[j];
      if (cip->ci_Status != RS_ACTIVE) continue;
      ++c;
      if (t == 0) {
         display_msg(0,"    Rsx  Last  Prev  Owner");
         t = 1;
         }
      sprintf(buf, "%2d  %2d  %5lu %5lu  %s",
         j+1,  /* 1-org */
         cip->ci_RuleSet,
         (now - cip->ci_LastTime)/100L, (now - cip->ci_PrevTime)/100L,
         cip->ci_Owner);
      display_msg(0,buf);
      }
   if (c == 0) display_msg(0,"No reader rows active");
   }

void init_monitor(int n_interfaces)
{
   int j;

#ifdef DOS
   set_tod();
   save_time();  /* For screen display */
#endif
   start_uptime_clock();  /* Starts 1 tick early; flows have CreateTime > 1 */

#if FLAT_MEM
   if ((fa = (struct flow huge *)flat_alloc(
#else  /* Other environments use malloc() */
   if ((fa = (struct flow huge *)farmalloc(
#endif
        sizeof(struct flow)*mxflows)) == NULL)
      log_msg(LOG_ERR, 1, "Not enough memory for flow table!");
   for (j = 0; j != mxflows; ++j) fa[j].FlowRuleSet = 0;

#if RS_CHAINS
   for (j = 0; j != mxflows-1; ++j) fa[j].rs_next = j+2;  /* 1-org */
   fa[mxflows-1].rs_next = 0;
   free_flow_head = 1;  free_flow_tail = mxflows;

   rsc_n_searches = 
      rsc_mx_range = rsc_av_range =
      rsc_mx_tests = rsc_av_tests = 0;
#endif

#if FLOW_INDEX
# if FLAT_MEM
   if ((flow_ix = (struct flow huge **)flat_alloc(
# else
   if ((flow_ix = (struct flow huge **)malloc(
# endif
         sizeof(struct flow huge *)*mxflows)) == NULL)
      log_msg(LOG_ERR, 2, "Not enough memory for flow index!");
   for (j = 0; j != mxflows; ++j) flow_ix[j] = NULL;
#endif

   nflows = 0;  /* No accounting flows in use yet */
   gcf_ix = empty_ix = mxflows;  /* Flow 1 will be the first allocated */

   for (j = sizeof(p2primes)/sizeof(Bit32)-1; 
      j != 8 && p2primes[j] > mxflows; --j) ;
   fthashmod = p2primes[j];  /* Min fthashmod = p2primes[8] (2039) */
#if FLAT_MEM
   flow_tbl_htp = (struct hash_tbl huge *)flat_alloc(
#else
   flow_tbl_htp = (struct hash_tbl huge *)farmalloc(
#endif
      sizeof(struct hash_tbl)+(fthashmod-1)*sizeof(struct flow huge *));
   if (flow_tbl_htp == NULL)
      log_msg(LOG_ERR, 3, "Not enough memory for flow table hash!");
   memset(flow_tbl_htp, 0, 
      sizeof(struct hash_tbl)+(fthashmod-1)*sizeof(struct flow huge *));
   t_hash_size = fthashmod;
   n_hash_tables = 1;

#if FLAT_MEM
   if ((rule_space = (struct rule huge *)flat_alloc(
#else
   if ((rule_space = (struct rule huge *)farmalloc(
#endif
         sizeof(struct rule)*mxrules)) == NULL)
      log_msg(LOG_ERR, 4, "Not enough memory for rulesets!");
   inuse_rules = 0;

   mxrulehash = 1 + mxrules*3/2;
#if FLAT_MEM
   if ((rule_hash_space = (Bit16 huge *)flat_alloc(
#else
   if ((rule_hash_space = (Bit16 huge *)farmalloc(
#endif
         sizeof(Bit16)*mxrulehash)) == NULL)
      log_msg(LOG_ERR, 5, "Not enough memory for rule hashes!");
   inuse_rule_hash = 0;

#if TCP_ATR
# if FLAT_MEM
   if ((sfa = (struct subflow huge *)flat_alloc(
# else
   if ((sfa = (struct subflow huge *)farmalloc(
# endif
        sizeof(struct subflow)*mxsubflows)) == NULL)
      log_msg(LOG_ERR, 6, "Not enough memory for subflows!");
   free_subflows = sfa;  n_free_subflows = mxsubflows;
   for (j = 0; j != mxsubflows-1; ++j) sfa[j].next_sf = &sfa[j+1];
   sfa[mxsubflows-1].next_sf = NULL;

   for (j = sizeof(p2primes)/sizeof(Bit32)-1; 
      j != 8 && p2primes[j] > mxsubflows; --j) ;
   sfhashmod = p2primes[j];
#if FLAT_MEM
   sfht = (struct subflow **)flat_alloc(
#else
   sfht = (struct subflow **)farmalloc(
#endif
      sfhashmod*sizeof(struct subflow *));
   if (sfht == NULL)
      log_msg(LOG_ERR, 7, "Not enough memory for subflow hash table!");
   memset(sfht, 0, sfhashmod*sizeof(struct subflow *));

# if FLAT_MEM
   if ((sfd = (struct subflow_data huge *)flat_alloc(
# else
   if ((sfd = (struct subflow_data huge *)farmalloc(
# endif
        sizeof(struct subflow_data)*mxsfd_blocks)) == NULL)
      log_msg(LOG_ERR, 8, "Not enough memory for subflow data blocks!");
   free_sfd_blocks = sfd;  n_free_sfd_blocks = mxsfd_blocks;
   for (j = 0; j != mxsfd_blocks-1; ++j) sfd[j].next = &sfd[j+1];
   sfd[mxsfd_blocks-1].next = NULL;

# if TCP_PKT_TRACE
   pt_file = fopen(ptf_title, "a");
# endif
#endif  /* TCP_ATR */

#if NEW_ATR
# if FLAT_MEM
   if ((distribs = (struct distribution huge *)flat_alloc(
# else
   if ((distribs = (struct distribution huge *)farmalloc(
# endif
        sizeof(struct distribution)*mxdistribs)) == NULL)
      log_msg(LOG_ERR, 9, "Not enough memory for distributions!");
   free_distribs = distribs;  n_free_distribs = mxdistribs;
   for (j = 0; j != mxdistribs-1; ++j) distribs[j].next = &distribs[j+1];
   distribs[mxdistribs-1].next = NULL;

# if FLAT_MEM
   if ((events = (struct event huge *)flat_alloc(
# else
   if ((events = (struct event huge *)farmalloc(
# endif
        sizeof(struct event)*mxdistevents)) == NULL)
      log_msg(LOG_ERR, 10, "Not enough memory for distrib events!");
   free_events = events;  n_free_events = mxdistevents;
   for (j = 0; j != mxdistevents-1; ++j) events[j].next = &events[j+1];
   events[mxdistevents-1].next = NULL;
#endif  /* NEW_ATR */

   memset(ri, 0, sizeof(ri));
   /* Create a ruleset info entry for the default rule table */
   ri[0].ri_rule_table = (struct rule huge *)default_rule_table;
   ri[0].ri_Size = DRT_SIZE;
   optimise_rule_table(1);  /* Alloc rule hash */
   memcpy(ri[0].ri_Owner, "NeTraMet", 8);
   memcpy(ri[0].ri_Name, "1", 1);  /* For compatibility with v3 */
   ri[0].ri_TimeStamp = 1;
   ri[0].ri_Status = RS_ACTIVE;

   tbl_size[2] = n_interfaces;  /* Interface table */

   memset(mi, 0, sizeof(mi));  /* Manager info */
   for (j = 1; j != sizeof(mi)/sizeof(struct mgr_rec); ++j)
      mi[j].mi_RunningStandby = TV_FALSE;
   memcpy(mi[0].mi_Owner, "NeTraMet", 8);
   mi[0].mi_CurrentRuleSet = 1;
   mi[0].mi_RunningStandby = TV_FALSE;
   mi[0].mi_TimeStamp = 1;
   mi[0].mi_Status = RS_ACTIVE;

   for (j = 0; j != sizeof(proto_reqd)/sizeof(int); ++j) proto_reqd[j] = 0;
   adj_reqd = 0;
   open_rule_set(1, 1); /* Open set 1 */

   memset(ci, 0, sizeof(ci));  /* Reader (collector) info */
   }

void show_help(void)
{
   display_msg(0,"Copyright (C) 1992-1999");
   display_msg(0,"  Nevil Brownlee");
   display_msg(0,"  ITSS Technology Development");
   display_msg(0,"  The University of Auckland");
   display_msg(0,"");
   display_msg(0,"Keyboard commands ..");
   display_msg(0,"");
   display_msg(0,"  e: show mEter reader table");
   display_msg(0,"  a: show mAnager table");
   display_msg(0,"  u: show rUleset table");
   display_msg(0,"  f: show Flow Table");
   display_msg(0,"  p: show Protocols");
   display_msg(0,"  s: show Statistics");
   display_msg(0,"  v: show meter Version");
#if defined(DOS)
   display_msg(0,"  z: Zero statistics");
# if !(OCX_NTM || OC3_NTM)
   display_msg(0,"  b: show Bad packet counts");
# endif
   display_msg(0,"  m: show Memory usage");
#endif
   if (kb_enabled) {
      display_msg(0,"");
      display_msg(0,"Esc: stop metering, exit NeTraMet");
      }
   }

void handle_kb(int ch)
{
   int c;
   switch (tolower(ch)) {
   case 'a':
      show_managers();
      break;
#if RS_CHAINS
   case 'C':
   case 'c':
      show_chains(ch == 'C');
      break;
#endif
   case 'e':
      show_readers();
      break;
   case 'f':
      show_flow_table();
      break;
   case 'm':
      show_memory();
      break;
   case 'p':
      show_protocols();
      break;
   case 'u':
      show_rulesets();
      break;
   case 's':
      show_stats();
      break;
   case 't':
      show_meter_time();
      break;
#if PM_DEBUG
   case '\\':
      printf("pm_debug: ");
#ifdef DOS
      c = getch();
#else
      c = getchar();
#endif
      if (c == '\\') {
         pm_dbg = 0;
         printf("ri[] off");
         }
      else {
         c -= '0';
         pm_dbg = c == 0 ? 10 : c;
         printf("ri[%d]", pm_dbg);
         }
      break;
#endif
   case 'z':
      zero_stats();
      break;
   case '?':
      show_help();
      break;
   default:
      break;
      }
   }
